# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from enum import Enum

class Gender(Enum):
    FEMALE = "female"
    MALE = "male"
    SECRET = "secret"

class Figure(Enum):
    LADIES = "ladies"
    MENS = "mens"
    BOYS = "boys"

class BreastSize(Enum):
    M = "m"
    L = "l"
    S = "s"
    NONE = "none"
    SS = "ss"

class Unit(Enum):
    VS = "piapro"
    LN = "light_sound"
    MMJ = "idol"
    VBS = "street"
    WXS = "theme_park"
    NIIGO = "school_refusal"
    NONE = "none"
    ANY = "any"

class SupportUnitType(Enum):
    NONE = "none"
    FULL = "full"
    UNIT = "unit"

class CharacterType(Enum):
    GAME_CHARACTER = "game_character"
    MOB = "mob"
    OUTSIDE_CHARACTER = "outside_character"

class ActionSetType(Enum):
    NORMAL = "normal"
    LIMITED = "limited"

class AreaType(Enum):
    REALITY_WORLD = "reality_world"
    SPIRIT_WORLD = "spirit_world"

class ViewType(Enum):
    SIDE_VIEW = "side_view"
    QUARTER_VIEW = "quarter_view"

class CardRarityType(Enum):
    RARITY_1 = "rarity_1"
    RARITY_2 = "rarity_2"
    RARITY_3 = "rarity_3"
    RARITY_4 = "rarity_4"
    RARITY_BIRTHDAY = "rarity_birthday"

class CardAttr(Enum):
    COOL = "cool"
    HAPPY = "happy"
    MYSTERIOUS = "mysterious"
    CUTE = "cute"
    PURE = "pure"
    ANY = "any"

class ArchiveDisplayType(Enum):
    HIDE = "hide"
    NONE = "none"

class CardParameterType(Enum):
    PARAM1 = "param1"
    PARAM2 = "param2"
    PARAM3 = "param3"

class DescriptionSpriteName(Enum):
    SCORE_UP = "score_up"
    JUDGMENT_UP = "judgment_up"
    LIFE_RECOVERY = "life_recovery"

class CardEpisodePartType(Enum):
    FIRST_PART = "first_part"
    SECOND_PART = "second_part"

class MusicDifficultyType(Enum):
    EASY = "easy"
    NORMAL = "normal"
    HARD = "hard"
    EXPERT = "expert"
    MASTER = "master"

class MusicVocalType(Enum):
    ORIGINAL_SONG = "original_song"
    SEKAI = "sekai"
    VIRTUAL_SINGER = "virtual_singer"
    ANOTHER_VOCAL = "another_vocal"
    INSTRUMENTAL = "instrumental"
    APRIL_FOOL_2022 = "april_fool_2022"

class DefaultMusicType(Enum):
    ORIGINAL_MUSIC = "original_music"
    SEKAI = "sekai"

class MusicAchievementType(Enum):
    SCORE_RANK = "score_rank"
    COMBO = "combo"

class MusicAssetType(Enum):
    JACKET = "jacket"

class ReleaseConditionType(Enum):
    NONE = "none"
    USER_RANK = "user_rank"
    UNIT_RANK = "unit_rank"
    MASTER_RANK = "master_rank"
    SHOP = "shop"
    CARD_LEVEL = "card_level"
    MUSIC_DIFFICULTY_BETTER_PLAY = "music_difficulty_better_play"
    UNIT_STORY = "unit_story"
    CHARACTER_RANK = "character_rank"
    INHERIT_COMPLETE = "inherit_complete"
    EVENT_POINT = "event_point"
    EVENT_STORY = "event_story"
    VIRTUAL_LIVE_PARTICIPATION = "virtual_live_participation"
    RANK_MATCH = "rank_match"
    SPECIAL_STORY = "special_story"
    ACTION_SET = "action_set"
    BONDS_RANK = "bonds_rank"
    CHARACTER_LIVE_USAGE_LEADER = "character_live_usage_leader"
    CHALLENGE_LIVE_CHARACTER_FORCE_RELEASE = "challenge_live_character_force_release"

class LiveType(Enum):
    CHALLENGE_LIVE = "challenge_live"
    CHEERFUL_CARNIVAL_OPEN = "cheerful_carnival_open"
    CHEERFUL_CARNIVAL_PRIVATE = "cheerful_carnival_private"
    MULTI_OPEN = "multi_open"
    PRELIMINARY_TOURNAMENT = "preliminary_tournament"
    RANK_MATCH = "rank_match"
    SOLO = "solo"

class IngameNoteType(Enum):
    NORMAL = "normal"
    LONG = "long"
    LONG_MID = "long_mid"
    LONG_END = "long_end"
    LONG_END_FLICK = "long_end_flick"
    FLICK = "flick"
    NORMAL_CRITICAL = "normal_critical"
    LONG_CRITICAL = "long_critical"
    LONG_MID_CRITICAL = "long_mid_critical"
    LONG_END_CRITICAL = "long_end_critical"
    LONG_END_FLICK_CRITICAL = "long_end_flick_critical"
    FLICK_CRITICAL = "flick_critical"
    LONG_COMBO_AUTO = "long_combo_auto"

class IngameNoteJudgeType(Enum):
    PERFECT = "perfect"
    GREAT = "great"
    GOOD = "good"
    BAD = "bad"
    MISS = "miss"
    DEAD = "dead"
    AUTO = "auto"

class IngameCutinCharacterType(Enum):
    NORMAL = "normal"
    BONDS = "bonds"

class ShopType(Enum):
    MUSIC = "music"
    LIVE_COSTUME = "live_costume"
    MUSIC_VOCAL = "music_vocal"
    AREA_ITEM = "area_item"
    STAMP = "stamp"

class MaterialType(Enum):
    COMMON = "common"
    COSTUME = "costume"
    MUSIC = "music"
    SPECIAL_TRAINING = "special_training"
    MASTER_LESSON = "master_lesson"
    CARD_TICKET = "card_ticket"
    GACHA_CEIL_TICKET = "gacha_ceil_ticket"
    VOCAL_CARD_TICKET = "vocal_card_ticket"

class GachaType(Enum):
    CEIL = "ceil"
    NORMAL = "normal"
    BEGINNER = "beginner"
    GIFT = "gift"

class ResourceType(Enum):
    PAID_JEWEL = "paid_jewel"
    COLORFUL_PASS = "colorful_pass"
    LIVE_POINT = "live_point"
    GACHA_TICKET = "gacha_ticket"
    MATERIAL = "material"
    PRACTICE_TICKET = "practice_ticket"
    COIN = "coin"
    COSTUME_3D = "costume_3d"
    VIRTUAL_COIN = "virtual_coin"
    JEWEL = "jewel"
    SKILL_PRACTICE_TICKET = "skill_practice_ticket"
    BOOST_ITEM = "boost_item"
    BONDS_HONOR = "bonds_honor"
    BONDS_HONOR_WORD = "bonds_honor_word"
    HONOR = "honor"
    AVATAR_COSTUME = "avatar_costume"
    STAMP = "stamp"
    CUSTOM_PROFILE_COLLECTION_RESOURCE = "custom_profile_collection_resource"
    CARD = "card"
    GACHA_CEIL_ITEM = "gacha_ceil_item"
    MUSIC = "music"
    MUSIC_VOCAL = "music_vocal"
    AVATAR_MOTION = "avatar_motion"
    AVATAR_SKIN_COLOR = "avatar_skin_color"
    AVATAR_COORDINATE = "avatar_coordinate"
    CHARACTER_RANK_EXP = "character_rank_exp"
    AREA_ITEM = "area_item"
    AVATAR_ACCESSORY = "avatar_accessory"
    PENLIGHT = "penlight"
    VIRTUAL_LIVE_TICKET = "virtual_live_ticket"
    VIRTUAL_LIVE_PAMPHLET = "virtual_live_pamphlet"
    EVENT_ITEM = "event_item"

class LevelType(Enum):
    CARD_SKILL_1 = "card_skill_1"
    CARD_SKILL_2 = "card_skill_2"
    CARD_SKILL_3 = "card_skill_3"
    CARD_SKILL_4 = "card_skill_4"
    CARD = "card"
    USER = "user"
    UNIT = "unit"
    CHARACTER = "character"
    CARD_SKILL_BIRTHDAY = "card_skill_birthday"
    BONDS = "bonds"

class ClientConfigType(Enum):
    INT = "Int"
    FLOAT = "Float"
    STRING = "String"

class Costume3dType(Enum):
    DEFAULT = "default"
    NORMAL = "normal"
    DISTRIBUTION = "distribution"

class PartType(Enum):
    HEAD = "head"
    BODY = "body"
    HAIR = "hair"

class Costume3dRarity(Enum):
    NORMAL = "normal"
    RARE = "rare"

class HeadCostume3dAssetbundleType(Enum):
    HEAD_ONLY = "head_only"
    HEAD_AND_HAIR = "head_and_hair"
    HEAD_ALL = "head_all"
    HEAD_FRONT = "head_front"
    HEAD_BACK = "head_back"

class MotionType(Enum):
    COSTUME_SETTING = "costume_setting"

class TopicType(Enum):
    SPECIAL_STORY_EPISODE = "special_story_episode"
    UNIT_STORY_EPISODE = "unit_story_episode"
    MUSIC_DIFFICULTY = "music_difficulty"
    RULE_SLIDE = "rule_slide"
    VIRTUAL_LIVE_REWARD = "virtual_live_reward"
    CHALLENGE_LIVE = "challenge_live"
    ATTRIBUTE_AREA_ITEM_SHOP = "attribute_area_item_shop"
    FACILITY_CARD_SKILL = "facility_card_skill"
    APP_RATING = "app_rating"
    VIRTUAL_LIVE_TUTORIAL = "virtual_live_tutorial"
    EVENT_STORY_EPISODE = "event_story_episode"
    RANK_MATCH_UNLOCK = "rank_match_unlock"
    INHERIT_COMPLETE = "inherit_complete"
    CUT_IN_VOICE = "cut_in_voice"

class StampType(Enum):
    ILLUSTRATION = "illustration"
    TEXT = "text"
    CHEERFUL_CARNIVAL_MESSAGE = "cheerful_carnival_message"

class MatchingLogic(Enum):
    RANDOM = "random"
    POWER = "power"

class MultiLiveLobbyType(Enum):
    NORMAL = "normal"
    CHEERFUL_CARNIVAL = "cheerful_carnival"

class RefreshCycle(Enum):
    NONE = "none"
    MONTHLY = "monthly"

class ExchangeCategory(Enum):
    MASTER_PIECE = "master_piece"
    CARD_TICKET = "card_ticket"
    VOCAL_CARD_TICKET = "vocal_card_ticket"
    GACHA_SEAL = "gacha_seal"
    MASTER_CRYSTAL = "master_crystal"

class MaterialExchangeType(Enum):
    NORMAL = "normal"
    BEGINNER = "beginner"

class BillingShopItemType(Enum):
    JEWEL = "jewel"
    VALUE_SET = "value_set"
    COLORFUL_PASS = "colorful_pass"
    LIVE_MISSION_PASS = "live_mission_pass"
    COSTUME_3D = "costume_3d"

class BillableLimitType(Enum):
    UNLIMITED = "unlimited"
    COUNT = "count"
    LIVE_POINT = "live_point"

class BillableLimitResetIntervalType(Enum):
    NONE = "none"
    MONTHLY = "monthly"
    DAY = "day"

class PurchaseOption(Enum):
    EXCHANGE = "exchange"

class JewelBehaviorType(Enum):
    BOOST_RECOVERY = "boost_recovery"

class CharacterMissionType(Enum):
    PLAY_LIVE = "play_live"
    WAITING_ROOM = "waiting_room"
    COLLECT_COSTUME_3D = "collect_costume_3d"
    COLLECT_STAMP = "collect_stamp"
    READ_AREA_TALK = "read_area_talk"
    READ_CARD_EPISODE_FIRST = "read_card_episode_first"
    READ_CARD_EPISODE_SECOND = "read_card_episode_second"
    COLLECT_ANOTHER_VOCAL = "collect_another_vocal"
    AREA_ITEM_LEVEL_UP_CHARACTER = "area_item_level_up_character"
    AREA_ITEM_LEVEL_UP_UNIT = "area_item_level_up_unit"
    AREA_ITEM_LEVEL_UP_REALITY_WORLD = "area_item_level_up_reality_world"
    COLLECT_MEMBER = "collect_member"
    SKILL_LEVEL_UP_RARE = "skill_level_up_rare"
    SKILL_LEVEL_UP_STANDARD = "skill_level_up_standard"
    MASTER_RANK_UP_RARE = "master_rank_up_rare"
    MASTER_RANK_UP_STANDARD = "master_rank_up_standard"

class NormalMissionType(Enum):
    MAKE_RARE_COSTUME_3D = "make_rare_costume_3d"
    SKILL_LEVEL_2 = "skill_level_2"
    MASTER_RANK = "master_rank"
    MAKE_ANOTHER_COLOR_COSTUME_3D = "make_another_color_costume_3d"
    CHARACTER_RANK_3 = "character_rank_3"
    SET_HONOR = "set_honor"
    CLEAR_LIVE_ANOTHER_VOCAL = "clear_live_another_vocal"
    CLEAR_SOLO_CHALLENGE_LIVE = "clear_solo_challenge_live"
    READ_CHARACTER_PROFILE_EPISODE = "read_character_profile_episode"
    READ_CARD_EPISODE_FIRST = "read_card_episode_first"
    READ_CARD_EPISODE_SECOND = "read_card_episode_second"
    USE_VIRTUAL_ITEM = "use_virtual_item"
    BUY_AVATAR_SKIN = "buy_avatar_skin"
    INHERIT_PLATFORM = "inherit_platform"
    CLEAR_VIRTUAL_LIVE = "clear_virtual_live"
    LIVE_CLEAR = "live_clear"
    MAKE_FRIEND = "make_friend"

class BeginnerMissionType(Enum):
    READ_UNIT_STORY_THEME_PARK = "read_unit_story_theme_park"
    READ_UNIT_STORY_IDOL = "read_unit_story_idol"
    READ_UNIT_STORY_STREET = "read_unit_story_street"
    READ_UNIT_STORY_LIGHT_SOUND = "read_unit_story_light_sound"
    READ_UNIT_STORY_SCHOOL_REFUSAL = "read_unit_story_school_refusal"
    READ_UNIT_STORY_PIAPRO = "read_unit_story_piapro"
    CLEAR_MULTI_LIVE = "clear_multi_live"
    BUY_MUSIC = "buy_music"
    MAKE_COSTUME_3D = "make_costume_3d"
    BUY_AREA_ITEM = "buy_area_item"
    ACHIEVE_BEGINNER_MISSION = "achieve_beginner_mission"

class BeginnerMissionCategory(Enum):
    NORMAL = "normal"
    SPECIAL = "special"

class ResourceBoxType(Enum):
    EXPAND = "expand"
    COSTUME_3D = "costume_3d"
    LIST = "list"

class LiveMissionType(Enum):
    FREE = "free"
    PREMIUM = "premium"

class LiveTalkType(Enum):
    SUCCESS = "success"
    FAILED = "failed"

class HonorType(Enum):
    CHARACTER = "character"
    ACHIEVEMENT = "achievement"
    EVENT = "event"
    RANK_MATCH = "rank_match"

class HonorRarity(Enum):
    LOW = "low"
    MIDDLE = "middle"
    HIGH = "high"
    HIGHEST = "highest"

class HonorMissionType(Enum):
    CLEAR_LIVE = "clear_live"
    PLAYER_RANK = "player_rank"
    LOGIN_CONTINUED = "login_continued"
    LOGIN_TOTAL = "login_total"
    COLLECT_ANOTHER_VOCAL = "collect_another_vocal"
    UNIT_RANK_ALL = "unit_rank_all"
    AREA_ITEM_5_LEVEL = "area_item_5_level"
    CHARACTER_RANK_ALL = "character_rank_all"
    CLEAR_LIVE_DIFFICULTY_ALL = "clear_live_difficulty_all"
    MASTER_FULL_COMBO = "master_full_combo"
    MASTER_FULL_PERFECT = "master_full_perfect"
    CLEAR_LIVE_TARGET_LIFE = "clear_live_target_life"
    CLEAR_LIVE_CONTINUED = "clear_live_continued"
    CLEAR_LIVE_COMBO = "clear_live_combo"
    COLLECT_COIN_TOTAL = "collect_coin_total"
    CLEAR_MULTI_LIVE_FULL_COMBO_ALL = "clear_multi_live_full_combo_all"
    MULTI_LIVE_SELECT_SAME_MUSIC = "multi_live_select_same_music"
    MULTI_LIVE_SAME_HONOR = "multi_live_same_honor"
    MULTI_LIVE_STAMP = "multi_live_stamp"
    DUPLICATE_CARD_TOTAL = "duplicate_card_total"
    COLLECT_STAMP = "collect_stamp"
    ACTION_SET = "action_set"
    READ_STORY_NO_SKIP = "read_story_no_skip"
    CLEAR_LIVE_SCHOOL_YEAR_1 = "clear_live_school_year_1"
    CLEAR_LIVE_SCHOOL_YEAR_2 = "clear_live_school_year_2"
    CLEAR_LIVE_SCHOOL_MIYAMASUZAKA = "clear_live_school_miyamasuzaka"
    CLEAR_LIVE_SCHOOL_KAMIYAMA = "clear_live_school_kamiyama"
    CLEAR_LIVE_MALE = "clear_live_male"
    MULTI_LIVE_MVP = "multi_live_mvp"
    MULTI_LIVE_SUPER_STAR = "multi_live_super_star"
    COLLECT_COSTUME_3D = "collect_costume_3d"
    LIGHT_SOUND_VIRTUAL_LIVE_TOTAL_CHEER_POINT = "light_sound_virtual_live_total_cheer_point"
    IDOL_VIRTUAL_LIVE_TOTAL_CHEER_POINT = "idol_virtual_live_total_cheer_point"
    STREET_VIRTUAL_LIVE_TOTAL_CHEER_POINT = "street_virtual_live_total_cheer_point"
    THEME_PARK_VIRTUAL_LIVE_TOTAL_CHEER_POINT = "theme_park_virtual_live_total_cheer_point"
    SCHOOL_REFUSAL_VIRTUAL_LIVE_TOTAL_CHEER_POINT = "school_refusal_virtual_live_total_cheer_point"
    PIAPRO_VIRTUAL_LIVE_TOTAL_CHEER_POINT = "piapro_virtual_live_total_cheer_point"
    SEND_FRIEND_REQUEST = "send_friend_request"
    MAKE_FRIEND = "make_friend"
    CLEAR_MULTI_LIVE_ALL_FRIEND = "clear_multi_live_all_friend"

class BondsRewardType(Enum):
    RESOURCE = "resource"
    CUT_IN_VOICE = "cut_in_voice"

class VirtualLiveType(Enum):
    BEGINNER = "beginner"
    NORMAL = "normal"
    ARCHIVE = "archive"
    CHEERFUL_CARNIVAL = "cheerful_carnival"
    STREAMING = "streaming"

class VirtualShopType(Enum):
    OTHER = "other"
    EVENT = "event"
    BIRTHDAY = "birthday"
    SPECIAL_LIVE = "special_live"
    VIRTUAL_LIVE_TICKET = "virtual_live_ticket"

class VirtualItemCategory(Enum):
    NORMAL = "normal"
    SPREAD = "spread"

class EffectExpressionType(Enum):
    THROW_EFFECT = "throw_effect"

class MessageSize(Enum):
    SMALL = "small"
    MEDIUM = "medium"
    LARGE = "large"
    X_LARGE = "x_large"

class VirtualLiveTicketType(Enum):
    PAID = "paid"

class RuleSlideType(Enum):
    MUSIC_SHOP = "MUSIC_SHOP"
    WORLD_MAP = "WORLD_MAP"
    MULTI_LIVE = "MULTI_LIVE"
    COSTUME_CHANGE = "COSTUME_CHANGE"
    COSTUME_CHANGE_MIKU = "COSTUME_CHANGE_MIKU"
    AREA_SHOP_SEKAI = "AREA_SHOP_SEKAI"
    VIRTUAL_LIVE_OVERVIEW = "VIRTUAL_LIVE_OVERVIEW"
    VIRTUAL_WAITING_ROOM = "VIRTUAL_WAITING_ROOM"
    VIRTUAL_IN_LIVE = "VIRTUAL_IN_LIVE"
    COSTUME_SHOP = "COSTUME_SHOP"
    CHARACTER_RANK = "CHARACTER_RANK"
    CHALLENGE_LIVE_CHARACTER_SELECT = "CHALLENGE_LIVE_CHARACTER_SELECT"
    CHALLENGE_LIVE_CHARACTER_ADD = "CHALLENGE_LIVE_CHARACTER_ADD"
    SKILL_UP = "SKILL_UP"
    STAMP_SHOP = "STAMP_SHOP"
    MASTER_TRAINING = "MASTER_TRAINING"
    ANOTHER_VOCAL_SHOP = "ANOTHER_VOCAL_SHOP"
    WAITING_ROOM = "WAITING_ROOM"
    FREE_STAMP = "FREE_STAMP"
    AREA_SHOP_SCHOOL = "AREA_SHOP_SCHOOL"
    MISSION = "MISSION"
    SUMMARY = "SUMMARY"
    CHALLENGE_LIVE = "CHALLENGE_LIVE"
    FREE_LIVE = "FREE_LIVE"
    LESSON = "LESSON"
    SPECIAL_TRAINING = "SPECIAL_TRAINING"
    LIVE_BONUS = "LIVE_BONUS"
    EVENT_TOP = "EVENT_TOP"
    EVENT_DECK = "EVENT_DECK"
    GUILD_VIRTUAL_LIVE_TOP = "GUILD_VIRTUAL_LIVE_TOP"
    AUTO_LIVE = "AUTO_LIVE"
    VIRTUAL_LIVE_MESSAGE = "VIRTUAL_LIVE_MESSAGE"
    GACHA_ITEM = "GACHA_ITEM"
    STORY_EVENT = "STORY_EVENT"
    FRIEND_LIST = "FRIEND_LIST"
    SELECT_LIST = "SELECT_LIST"
    CHEERFUL_CARNIVAL = "CHEERFUL_CARNIVAL"
    GACHA_BONUS = "GACHA_BONUS"
    CHANGEROOM_MEMBER = "CHANGEROOM_MEMBER"
    CUSTOM_PROFILE = "CUSTOM_PROFILE"
    PIAPRO_BONUS = "PIAPRO_BONUS"
    BONDS = "BONDS"
    RANK_MATCH = "RANK_MATCH"
    GIFT_GACHA = "GIFT_GACHA"
    COLORFUL_SELECT_LIST = "COLORFUL_SELECT_LIST"
    STAMP_MISSION = "STAMP_MISSION"

class FacilityType(Enum):
    SKILL_PRACTICE = "skill_practice"
    CHALLENGE_LIVE = "challenge_live"
    RANK_MATCH = "rank_match"

class OneTimeBehaviorType(Enum):
    CHALLENGE_LIVE_CHARACTER_FORCE_RELEASE = "challenge_live_character_force_release"
    VIRTUAL_LIVE_TUTORIAL = "virtual_live_tutorial"

class EventType(Enum):
    MARATHON = "marathon"
    CHEERFUL_CARNIVAL = "cheerful_carnival"

class EventStoryUnitRelation(Enum):
    MAIN = "main"
    SUB = "sub"

class PreliminaryTournamentType(Enum):
    FIRST = "first"
    SECOND = "second"

class CheerfulCarnivalTeamPointTermType(Enum):
    CHEERFUL_CARNIVAL_TEAM_POINT_1ST = "cheerful_carnival_team_point_1st"
    CHEERFUL_CARNIVAL_TEAM_POINT_2ND = "cheerful_carnival_team_point_2nd"
    CHEERFUL_CARNIVAL_TEAM_POINT_FINAL = "cheerful_carnival_team_point_final"

class CheerfulCarnivalResultType(Enum):
    VICTORY = "victory"
    DEFEAT = "defeat"
    DRAW = "draw"

class AppealTargetType(Enum):
    GACHA = "gacha"

class AppealType(Enum):
    ONCE = "once"

class AppealReadConditionType(Enum):
    TRANSITION = "transition"

class StoryType(Enum):
    UNIT_STORY = "unit_story"
    EVENT_STORY = "event_story"
    CARD_STORY = "card_story"

class CustomProfileResourceType(Enum):
    PLAYER_INFO = "player_info"
    GENERAL_BACKGROUND = "general_background"
    STORY_BACKGROUND = "story_background"
    COLLECTION = "collection"
    MEMBER_STANDING_PICTURE = "member_standing_picture"
    SHAPE = "shape"
    ETC = "etc"

class ResourceLoadType(Enum):
    PREFAB = "prefab"
    ASSETBUNDLE = "assetbundle"

class CustomProfileResourceCollectionType(Enum):
    OMIKUJI = "omikuji"
    CAN_BADGE = "can_badge" 

class FortuneType(Enum):
    GRATE_FORTUNE = "grate_fortune"
    FORTUNE = "fortune"
    MIDDLE_FORTUNE = "middle_fortune"
    SMALL_FORTUNE = "small_fortune"
    UNCERTAIN_LUCK = "uncertain_luck"
    MISFORTUNE = "misfortune"
    GRATE_MISFORTUNE = "grate_misfortune"

class VirtualBoothShopType(Enum):
    VIRTUAL_SHOP = "virtual_shop"
    OMIKUJI = "omikuji"
    COLLECTION_GACHA = "collection_gacha"

class SpecialSeasonAreaUseType(Enum):
    BGM = "bgm"
    AREA = "area"

class RankMatchPenaltyType(Enum):
    WARNING_POP = "warning_pop"
    MINUTES_BAN = "minutes_ban"
    HOURS_BAN = "hours_ban"
    SEASON_BAN = "season_ban"

class RankMatchPlacementConditionType(Enum):
    ALL_PERFECT = "all_perfect"
    ALL_WIN = "all_win"
    ALL_LOSE = "all_lose"
    LIFE_0 = "life_0"

class TierBehaviorType(Enum):
    CLASS_UP = "class_up"
    CLASS_DOWN = "class_down"

class RankMatchBonusPointConditionType(Enum):
    ALL_PERFECT = "all_perfect"
    FULL_COMBO = "full_combo"
    HIGHER_CLASS_WIN = "higher_class_win"
    LIFE_0 = "life_0"

class CalcType(Enum):
    ADD = "add"
    SUBTRACT = "subtract"

class ActivateEffectValueType(Enum):
    RATE = "rate"
    FIXED = "fixed"

class SkillEffectType(Enum):
    SCORE_UP = "score_up"
    JUDGMENT_UP = "judgment_up"
    LIFE_RECOVERY = "life_recovery"
    SCORE_UP_CONDITION_LIFE = "score_up_condition_life"
    SCORE_UP_KEEP = "score_up_keep"

class SkillEffectConditionType(Enum):
    EQUALS_OR_OVER = "equals_or_over"

class SkillEnhanceType(Enum):
    SUB_UNIT_SCORE_UP = "sub_unit_score_up"

class MusicCategory(Enum):
    MV = "mv"
    MV_2D = "mv_2d"
    IMAGE = "image"
    ORIGINAL = "original"

class LotteryType(Enum):
    NORMAL = "normal"

class GachaBehaviorType(Enum):
    ONCE_A_DAY = "once_a_day"
    NORMAL = "normal"
    OVER_RARITY_3_ONCE = "over_rarity_3_once"
    OVER_RARITY_4_ONCE = "over_rarity_4_once"

class GachaPickupType(Enum):
    NORMAL = "normal"

class MissionType(Enum):
    NORMAL_MISSION = "normal_mission"
    BEGINNER_MISSION = "beginner_mission"
    LIVE_MISSION = "live_mission"
    HONOR_MISSION = "honor_mission"

class ResourceBoxPurpose(Enum):
    BILLING_SHOP_ITEM = "billing_shop_item"
    BILLING_SHOP_ITEM_BONUS = "billing_shop_item_bonus"
    BILLING_SHOP_ITEM_COUNT_BONUS = "billing_shop_item_count_bonus"
    BONDS_REWARD = "bonds_reward"
    CARD_EXCHANGE_RESOURCE = "card_exchange_resource"
    CHALLENGE_LIVE_HIGH_SCORE = "challenge_live_high_score"
    CHALLENGE_LIVE_PLAY_DAY_REWARD = "challenge_live_play_day_reward"
    CHALLENGE_LIVE_SCORE_RANK_REWARD_DETAIL = "challenge_live_score_rank_reward_detail"
    CHALLENGE_LIVE_STAGE = "challenge_live_stage"
    CHALLENGE_LIVE_STAGE_EX = "challenge_live_stage_ex"
    CHARACTER_RANK_REWARD = "character_rank_reward"
    CHEERFUL_CARNIVAL_RESULT_REWARD = "cheerful_carnival_result_reward"
    CHEERFUL_CARNIVAL_REWARD = "cheerful_carnival_reward"
    COLORFUL_PASS = "colorful_pass"
    COMPENSATION = "compensation"
    CONNECT_LIVE_REWARD = "connect_live_reward"
    CONVERT_GACHA_CEIL_ITEM = "convert_gacha_ceil_item"
    EPISODE_REWARD = "episode_reward"
    EVENT_EXCHANGE = "event_exchange"
    EVENT_RANKING_REWARD = "event_ranking_reward"
    GACHA_CEIL_EXCHANGE = "gacha_ceil_exchange"
    GACHA_EXTRA = "gacha_extra"
    GIFT_DETAIL = "gift_detail"
    GIFT_GACHA_EXTRA = "gift_gacha_extra"
    LOGIN_BONUS = "login_bonus"
    MASTER_LESSON_REWARD = "master_lesson_reward"
    MATERIAL_EXCHANGE = "material_exchange"
    MISSION_REWARD = "mission_reward"
    MULTI_SCORE_RANK_REWARD_DETAIL = "multi_score_rank_reward_detail"
    MUSIC_ACHIEVEMENT = "music_achievement"
    PLAYER_RANK_REWARD = "player_rank_reward"
    RANK_MATCH_SEASON_TIER_REWARD = "rank_match_season_tier_reward"
    SCORE_RANK_REWARD_DETAIL = "score_rank_reward_detail"
    SHOP_ITEM = "shop_item"
    SUPER_FEVER_REWARD = "super_fever_reward"
    VIRTUAL_LIVE_CHEER_POINT_REWARD = "virtual_live_cheer_point_reward"
    VIRTUAL_LIVE_MEMBER_COUNT_REWARD = "virtual_live_member_count_reward"
    VIRTUAL_LIVE_REWARD = "virtual_live_reward"
    VIRTUAL_SHOP_ITEM = "virtual_shop_item"

class GachaCeilExchangeLabelType(Enum):
    FES = "fes"
    LIMITED = "limited"

class VirtualLiveSetlistType(Enum):
    MC = "mc"
    MUSIC = "music"

class DayOfWeek(Enum):
    MONDAY = "monday"
    TUESDAY = "tuesday"
    WEDNESDAY = "wednesday"
    THURSDAY = "thursday"
    FRIDAY = "friday"
    SATURDAY = "saturday"
    SUNDAY = "sunday"

class VirtualItemLabelType(Enum):
    SPECIAL = "special"

class VirtualLiveStageStatus(Enum):
    OPEN = "open"
    LIVE = "live"

class VirtualShopItemType(Enum):
    AVATAR_MOTION = "avatar_motion"
    COSTUME = "costume"
    ACCESSORY = "accessory"
    COORDINATE = "coordinate"
    PENLIGHT = "penlight"
    VIRTUAL_LIVE_TICKET = "virtual_live_ticket"
    BOOST_ITEM = "boost_item"
    PREMIUM_VIRTUAL_LIVE_TICKET = "premium_virtual_live_ticket"

class AccessoryPart(Enum):
    ACCESSORY_FACE = "Accessory_face"

class SpecialSeasonType(Enum):
    APRIL_FOOL_2022 = "april_fool_2022"

class AppVersionStatus(Enum):
    NOT_AVAILABLE = "not_available"
    AVAILABLE = "available"
    MAINTENANCE = "maintenance"

class BundleCategory(Enum):
    STARTAPP = "StartApp"
    ONDEMAND = "OnDemand"
    ADDITIONALVOICE = "AdditionalVoice"
    TUTORIAL = "Tutorial"



# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from pydantic import BaseModel, Extra

class ModelWithExtra(BaseModel):
    class Config:
        extra = Extra.allow


# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from typing import Iterator, Optional, List
from msgpack import packb, unpackb
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad

def msgpack(dict: Optional[dict]) -> bytes:
    return b"" if dict is None else packb(dict) 

def unmsgpack(data: bytes) -> dict:
    return unpackb(data, strict_map_key=False) if len(data)>0 else None

def encrypt(plaintext: bytes, key: bytes, iv: bytes) -> bytes:
    cipher = AES.new(key, AES.MODE_CBC, iv=iv)
    ciphertext: bytes = cipher.encrypt(pad(plaintext, 16))
    return ciphertext

def decrypt(ciphertext: bytes, key: bytes, iv: bytes) -> bytes:
    cipher = AES.new(key, AES.MODE_CBC, iv=iv)
    plaintext: bytes = unpad(cipher.decrypt(ciphertext), 16)
    return plaintext

def deobfuscated(obfuscatedChunks: Iterator[List[bytes]]) -> Iterator:
    return (chunk if chunkIndex>0 else bytes([byte[0] if i>=128 or i%8>=5 else byte[0]^0xFF for i,byte in enumerate(chunk[4:])]) for chunkIndex, chunk in enumerate(obfuscatedChunks))

def obfuscated(chunks: Iterator[List[bytes]]) -> Iterator:
    return (chunk if chunkIndex>0 else bytes([0x10,0,0,0]+[byte[0] if i>=128 or i%8>=5 else byte[0]^0xFF for i,byte in enumerate(chunk)]) for chunkIndex, chunk in enumerate(chunks))



# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from typing import Iterator, List, Optional

from pjsekai.utilities import deobfuscated, obfuscated

class AssetBundle:

    _chunks: Iterator[List[bytes]]
    _obfuscatedChunks: Iterator[List[bytes]]
    
    @property
    def chunks(self) -> Iterator[List[bytes]]:
        return self._chunks
    @property
    def obfuscatedChunks(self) -> Iterator[List[bytes]]:
        return self._obfuscatedChunks

    def __init__(self, chunks: Optional[Iterator[List[bytes]]] = None, obfuscatedChunks: Optional[Iterator[List[bytes]]] = None):
        if chunks is not None:
            self._chunks = chunks
            self._obfuscatedChunks = obfuscated(chunks)
        elif obfuscatedChunks is not None:
            self._chunks = deobfuscated(obfuscatedChunks)
            self._obfuscatedChunks = obfuscatedChunks
        else:
            raise ValueError
            
    def extract(self) -> None:
        raise NotImplementedError



# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from datetime import datetime
from typing import List, Optional, Union

from .model import ModelWithExtra
from pjsekai.enums import *


class GameCharacter(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    resourceId: Optional[int]
    firstName: Optional[str]
    givenName: Optional[str]
    firstNameRuby: Optional[str]
    givenNameRuby: Optional[str]
    gender: Optional[Union[Gender, Unknown]]
    height: Optional[float]
    live2dHeightAdjustment: Optional[float]
    figure: Optional[Union[Figure, Unknown]]
    breastSize: Optional[Union[BreastSize, Unknown]]
    modelName: Optional[str]
    unit: Optional[Union[Unit, Unknown]]
    supportUnitType: Optional[Union[SupportUnitType, Unknown]]


class GameCharacterUnit(ModelWithExtra):
    id: Optional[int]
    gameCharacterId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    colorCode: Optional[str]
    skinColorCode: Optional[str]
    skinShadowColorCode1: Optional[str]
    skinShadowColorCode2: Optional[str]


class OutsideCharacter(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]


class Character3d(ModelWithExtra):
    id: Optional[int]
    characterType: Optional[Union[CharacterType, Unknown]]
    characterId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    name: Optional[str]
    headCostume3dId: Optional[int]
    hairCostume3dId: Optional[int]
    bodyCostume3dId: Optional[int]


class Character2d(ModelWithExtra):
    id: Optional[int]
    characterType: Optional[Union[CharacterType, Unknown]]
    characterId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    assetName: Optional[str]


class CharacterProfile(ModelWithExtra):
    characterId: Optional[int]
    characterVoice: Optional[str]
    birthday: Optional[str]
    height: Optional[str]
    school: Optional[str]
    schoolYear: Optional[str]
    hobby: Optional[str]
    specialSkill: Optional[str]
    favoriteFood: Optional[str]
    hatedFood: Optional[str]
    weak: Optional[str]
    introduction: Optional[str]
    scenarioId: Optional[str]


class Bond(ModelWithExtra):
    id: Optional[int]
    groupId: Optional[int]
    characterId1: Optional[int]
    characterId2: Optional[int]


class Live2d(ModelWithExtra):
    id: Optional[int]
    characterId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    assetbundleName: Optional[str]
    motion: Optional[str]
    expression: Optional[str]
    weight: Optional[int]


class BondsLive2d(Live2d):
    defaultFlg: Optional[bool]


class BondsRankUpLive2d(Live2d):
    defaultFlg: Optional[bool]


class UnitProfile(ModelWithExtra):
    unit: Optional[Union[Unit, Unknown]]
    unitName: Optional[str]
    seq: Optional[int]
    profileSentence: Optional[str]
    colorCode: Optional[str]


class ActionSet(ModelWithExtra):
    id: Optional[int]
    areaId: Optional[int]
    scriptId: Optional[str]
    characterIds: Optional[List[int]]
    archiveDisplayType: Optional[Union[ArchiveDisplayType, Unknown]]
    archivePublishedAt: Optional[datetime]
    releaseConditionId: Optional[datetime]
    scenarioId: Optional[str]
    actionSetType: Optional[Union[ActionSetType, Unknown]]
    specialSeasonId: Optional[int]


class Area(ModelWithExtra):
    id: Optional[int]
    assetbundleName: Optional[str]
    areaType: Optional[Union[AreaType, Unknown]]
    viewType: Optional[Union[ViewType, Unknown]]
    name: Optional[str]
    releaseConditionId: Optional[int]
    label: Optional[str]
    startAt: Optional[datetime]
    endAt: Optional[datetime]


class AreaPlaylist(ModelWithExtra):
    id: Optional[int]
    areaId: Optional[int]
    musicId: Optional[int]
    assetbundleName: Optional[str]
    bgmName: Optional[str]
    releaseConditionId: Optional[int]


class MobCharacter(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    gender: Optional[Union[Gender, Unknown]]


class CharacterCostume(ModelWithExtra):
    id: Optional[int]
    characterId: Optional[int]
    costumeId: Optional[int]
    sdAssetbundleName: Optional[str]
    live2dAssetbundleName: Optional[str]


class CardCostume3d(ModelWithExtra):
    cardId: Optional[int]
    costume3dId: Optional[int]


class CardParameter(ModelWithExtra):
    id: Optional[int]
    cardId: Optional[int]
    cardLevel: Optional[int]
    cardParameterType: Optional[Union[CardParameterType, Unknown]]
    power: Optional[int]


class Cost(ModelWithExtra):
    resourceId: Optional[int]
    resourceType: Optional[Union[ResourceType, Unknown]]
    resourceLevel: Optional[int]
    quantity: Optional[int]


class SpecialTrainingCost(ModelWithExtra):
    cardId: Optional[int]
    seq: Optional[int]
    cost: Optional[Cost]


class MasterLessonAchieveResource(ModelWithExtra):
    releaseConditionId: Optional[int]
    cardId: Optional[int]
    masterRank: Optional[int]
    resources: Optional[List[Union[dict, str, int]]]


class Card(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    characterId: Optional[int]
    cardRarityType: Optional[Union[CardRarityType, Unknown]]
    specialTrainingPower1BonusFixed: Optional[int]
    specialTrainingPower2BonusFixed: Optional[int]
    specialTrainingPower3BonusFixed: Optional[int]
    attr: Optional[Union[CardAttr, Unknown]]
    supportunit: Optional[Union[Unit, Unknown]]
    skillId: Optional[int]
    cardSkillName: Optional[str]
    prefix: Optional[str]
    assetbundleName: Optional[str]
    gachaPhrase: Optional[str]
    flavorText: Optional[str]
    releaseAt: Optional[datetime]
    archivePublishedAt: Optional[datetime]
    cardParameters: Optional[List[CardParameter]]
    specialTrainingCosts: Optional[List[SpecialTrainingCost]]
    masterLessonAchieveResources: Optional[List[MasterLessonAchieveResource]]
    archiveDisplayType: Optional[Union[ArchiveDisplayType, Unknown]]


class SkillEffectDetail(ModelWithExtra):
    id: Optional[int]
    level: Optional[int]
    activateEffectDuration: Optional[float]
    activateEffectValueType: Optional[Union[ActivateEffectValueType, Unknown]]
    activateEffectValue: Optional[int]


class SkillEnhanceCondition(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    unit: Optional[Union[Unit, Unknown]]


class SkillEnhance(ModelWithExtra):
    id: Optional[int]
    skillEnhanceType: Optional[Union[SkillEnhanceType, Unknown]]
    activateEffectValueType: Optional[Union[ActivateEffectValueType, Unknown]]
    activateEffectValue: Optional[int]
    skillEnhanceCondition: Optional[SkillEnhanceCondition]


class SkillEffect(ModelWithExtra):
    id: Optional[int]
    skillEffectType: Optional[Union[SkillEffectType, Unknown]]
    activateNotesJudgmentType: Optional[Union[IngameNoteJudgeType, Unknown]]
    skillEffectDetails: Optional[List[SkillEffectDetail]]
    activateLife: Optional[int]
    conditionType: Optional[Union[SkillEffectConditionType, Unknown]]
    skillEnhance: Optional[SkillEnhance]


class Skill(ModelWithExtra):
    id: Optional[int]
    shortDescription: Optional[str]
    description: Optional[str]
    descriptionSpriteName: Optional[str]
    skillFilterId: Optional[int]
    skillEffects: Optional[List[SkillEffect]]


class CardEpisode(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    cardId: Optional[int]
    title: Optional[str]
    scenarioId: Optional[str]
    assetbundleName: Optional[str]
    releaseConditionId: Optional[int]
    power1BonusFixed: Optional[int]
    power2BonusFixed: Optional[int]
    power3BonusFixed: Optional[int]
    rewardResourceBoxIds: Optional[List[int]]
    costs: Optional[List[Cost]]
    cardEpisodePartType: Optional[Union[CardEpisodePartType, Unknown]]


class CardRarity(ModelWithExtra):
    cardRarityType: Optional[Union[CardRarityType, Unknown]]
    seq: Optional[int]
    maxLevel: Optional[int]
    maxSkillLevel: Optional[int]
    trainingMaxLevel: Optional[int]


class CardSkillCost(ModelWithExtra):
    id: Optional[int]
    materialId: Optional[int]
    exp: Optional[int]


class Music(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    releaseConditionId: Optional[int]
    categories: Optional[List[Union[MusicCategory, Unknown]]]
    title: Optional[str]
    pronunciation: Optional[str]
    lyricist: Optional[str]
    composer: Optional[str]
    arranger: Optional[str]
    dancerCount: Optional[int]
    selfDancerPosition: Optional[int]
    assetbundleName: Optional[str]
    liveTalkBackgroundAssetbundleName: Optional[str]
    publishedAt: Optional[datetime]
    liveStageId: Optional[int]
    fillerSec: Optional[float]
    musicCollaborationId: Optional[int]


class MusicTag(ModelWithExtra):
    id: Optional[int]
    musicId: Optional[int]
    musicTag: Optional[str]
    seq: Optional[int]


class MusicDifficulty(ModelWithExtra):
    id: Optional[int]
    musicId: Optional[int]
    musicDifficulty: Optional[Union[MusicDifficultyType, Unknown]]
    playLevel: Optional[int]
    releaseConditionId: Optional[int]
    noteCount: Optional[int]


class Character(ModelWithExtra):
    id: Optional[int]
    musicId: Optional[int]
    musicVocalId: Optional[int]
    characterType: Optional[Union[CharacterType, Unknown]]
    characterId: Optional[int]
    seq: Optional[int]


class MusicVocal(ModelWithExtra):
    id: Optional[int]
    musicId: Optional[int]
    musicVocalType: Optional[Union[MusicVocalType, Unknown]]
    seq: Optional[int]
    releaseConditionId: Optional[int]
    caption: Optional[str]
    characters: Optional[List[Character]]
    assetbundleName: Optional[str]
    archivePublishedAt: Optional[datetime]
    specialSeasonId: Optional[int]
    archiveDisplayType: Optional[Union[ArchiveDisplayType, Unknown]]


class MusicDanceMember(ModelWithExtra):
    id: Optional[int]
    musicId: Optional[int]
    defaultMusicType: Optional[Union[DefaultMusicType, Unknown]]
    characterId1: Optional[int]
    unit1: Optional[Union[Unit, Unknown]]
    characterId2: Optional[int]
    unit2: Optional[Union[Unit, Unknown]]
    characterId3: Optional[int]
    unit3: Optional[Union[Unit, Unknown]]
    characterId4: Optional[int]
    unit4: Optional[Union[Unit, Unknown]]
    characterId5: Optional[int]
    unit5: Optional[Union[Unit, Unknown]]


class MusicAchievement(ModelWithExtra):
    id: Optional[int]
    musicAchievementType: Optional[Union[MusicAchievementType, Unknown]]
    musicAchievementTypeValue: Optional[str]
    resourceBoxId: Optional[int]
    musicDifficultyType: Optional[Union[MusicDifficultyType, Unknown]]


class MusicVideoCharacter(ModelWithExtra):
    id: Optional[int]
    musicId: Optional[int]
    defaultMusicType: Optional[Union[DefaultMusicType, Unknown]]
    gameCharacterUnitId: Optional[int]
    dancePriority: Optional[int]
    seq: Optional[int]
    priority: Optional[int]


class MusicAssetVariant(ModelWithExtra):
    id: Optional[int]
    musicVocalId: Optional[int]
    seq: Optional[int]
    musicAssetType: Optional[Union[MusicAssetType, Unknown]]
    assetbundleName: Optional[str]


class MusicCollaboration(ModelWithExtra):
    id: Optional[int]
    label: Optional[str]


class EpisodeMusicVideoCostume(ModelWithExtra):
    id: Optional[int]
    musicVocalId: Optional[int]
    character3dId1: Optional[int]
    character3dId2: Optional[int]
    character3dId3: Optional[int]
    character3dId4: Optional[int]
    character3dId5: Optional[int]


class ReleaseCondition(ModelWithExtra):
    id: Optional[int]
    sentence: Optional[str]
    releaseConditionType: Optional[Union[ReleaseConditionType, Unknown]]
    releaseConditionTypeLevel: Optional[int]
    releaseConditionTypeId: Optional[int]
    releaseConditionTypeQuantity: Optional[int]


class PlayLevelScore(ModelWithExtra):
    liveType: Optional[Union[LiveType, Unknown]]
    playLevel: Optional[int]
    s: Optional[int]
    a: Optional[int]
    b: Optional[int]
    c: Optional[int]


class IngameCombo(ModelWithExtra):
    id: Optional[int]
    fromCount: Optional[int]
    toCount: Optional[int]
    scoreCoefficient: Optional[float]


class IngameNote(ModelWithExtra):
    id: Optional[int]
    ingameNoteType: Optional[Union[IngameNoteType, Unknown]]
    scoreCoefficient: Optional[float]
    damageBad: Optional[int]
    damageMiss: Optional[int]


class IngameNoteJudge(ModelWithExtra):
    id: Optional[int]
    ingameNoteJadgeType: Optional[Union[IngameNoteJudgeType, Unknown]]
    scoreCoefficient: Optional[float]
    damage: Optional[int]


class IngamePlayLevel(ModelWithExtra):
    playLevel: Optional[int]
    scoreCoefficient: Optional[float]


class IngameCutin(ModelWithExtra):
    id: Optional[int]
    musicDifficulty: Optional[Union[MusicDifficultyType, Unknown]]
    combo: Optional[int]


class IngameCutinCharacter(ModelWithExtra):
    id: Optional[int]
    ingameCutinCharacterType: Optional[Union[IngameCutinCharacterType, Unknown]]
    priority: Optional[int]
    gameCharacterUnitId1: Optional[int]
    gameCharacterUnitId2: Optional[int]
    assetbundleName1: Optional[str]
    assetbundleName2: Optional[str]
    releaseConditionId: Optional[int]


class IngameJudgeFrame(ModelWithExtra):
    id: Optional[int]
    ingameNoteType: Optional[Union[IngameNoteType, Unknown]]
    perfect: Optional[float]
    great: Optional[float]
    good: Optional[float]
    bad: Optional[float]
    perfectBefore: Optional[float]
    perfectAfter: Optional[float]
    greatBefore: Optional[float]
    greatAfter: Optional[float]
    goodBefore: Optional[float]
    goodAfter: Optional[float]
    badBefore: Optional[float]
    badAfter: Optional[float]


class IngameNoteJudgeTechnicalScore(ModelWithExtra):
    id: Optional[int]
    liveType: Optional[Union[LiveType, Unknown]]
    ingameNoteJadgeType: Optional[Union[IngameNoteJudgeType, Unknown]]
    score: Optional[int]


class Shop(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    shopType: Optional[Union[ShopType, Unknown]]
    areaId: Optional[int]
    name: Optional[str]
    releaseConditionId: Optional[int]


class ShopItemCost(ModelWithExtra):
    shopItemId: Optional[int]
    seq: Optional[int]
    cost: Optional[Cost]


class ShopItem(ModelWithExtra):
    id: Optional[int]
    shopId: Optional[int]
    seq: Optional[int]
    releaseConditionId: Optional[int]
    resourceBoxId: Optional[int]
    costs: Optional[List[ShopItemCost]]
    startAt: Optional[datetime]


class Costume3dShopItemCost(Cost):
    id: Optional[int]
    costume3dShopItemId: Optional[int]
    seq: Optional[int]


class Costume3dShopItem(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    groupId: Optional[int]
    name: Optional[str]
    bodyCostume3dId: Optional[int]
    releaseConditionId: Optional[int]
    costs: Optional[List[Costume3dShopItemCost]]
    startAt: Optional[datetime]
    headCostume3dId: Optional[int]


class AreaItem(ModelWithExtra):
    id: Optional[int]
    areaId: Optional[int]
    name: Optional[str]
    flavorText: Optional[str]
    spawnPoint: Optional[str]
    assetbundleName: Optional[str]


class AreaItemLevel(ModelWithExtra):
    areaItemId: Optional[int]
    level: Optional[int]
    targetunit: Optional[Union[Unit, Unknown]]
    targetCardAttr: Optional[Union[CardAttr, Unknown]]
    targetGameCharacterId: Optional[int]
    power1BonusRate: Optional[float]
    power1AllMatchBonusRate: Optional[float]
    power2BonusRate: Optional[float]
    power2AllMatchBonusRate: Optional[float]
    power3BonusRate: Optional[float]
    power3AllMatchBonusRate: Optional[float]
    sentence: Optional[str]


class Material(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    flavorText: Optional[str]
    materialType: Optional[Union[MaterialType, Unknown]]


class GachaCardRarityRate(ModelWithExtra):
    id: Optional[int]
    groupId: Optional[int]
    cardRarityType: Optional[Union[CardRarityType, Unknown]]
    lotteryType: Optional[Union[LotteryType, Unknown]]
    rate: Optional[float]


class GachaDetail(ModelWithExtra):
    id: Optional[int]
    gachaId: Optional[int]
    cardId: Optional[int]
    weight: Optional[int]
    isWish: Optional[bool]


class GachaBehavior(ModelWithExtra):
    id: Optional[int]
    gachaId: Optional[int]
    groupId: Optional[int]
    priority: Optional[int]
    gachaBehaviorType: Optional[Union[GachaBehaviorType, Unknown]]
    costResourceType: Optional[Union[ResourceType, Unknown]]
    costResourceQuantity: Optional[int]
    spinCount: Optional[int]
    executeLimit: Optional[int]
    costResourceId: Optional[int]
    gachaExtraId: Optional[int]


class GachaPickup(ModelWithExtra):
    id: Optional[int]
    gachaId: Optional[int]
    cardId: Optional[int]
    gachaPickupType: Optional[Union[GachaPickupType, Unknown]]


class GachaInformation(ModelWithExtra):
    gachaId: Optional[int]
    summary: Optional[str]
    description: Optional[str]


class Gacha(ModelWithExtra):
    id: Optional[int]
    gachaType: Optional[Union[GachaType, Unknown]]
    name: Optional[str]
    seq: Optional[int]
    assetbundleName: Optional[str]
    gachaCardRarityRateGroupId: Optional[int]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    isShowPeriod: Optional[bool]
    gachaCeilItemId: Optional[int]
    wishSelectCount: Optional[int]
    wishFixedSelectCount: Optional[int]
    wishLimitedSelectCount: Optional[int]
    gachaCardRarityRates: Optional[List[GachaCardRarityRate]]
    gachaDetails: Optional[List[GachaDetail]]
    gachaBehaviors: Optional[List[GachaBehavior]]
    gachaPickups: Optional[List[GachaPickup]]
    gachaPickupCostumes: Optional[List[Union[dict, str, int]]]
    gachaInformation: Optional[GachaInformation]
    drawableGachaHour: Optional[int]
    gachaBonusId: Optional[int]
    spinLimit: Optional[int]


class GachaBonus(ModelWithExtra):
    id: Optional[int]


class GachaBonusPoint(ModelWithExtra):
    id: Optional[int]
    resourceType: Optional[Union[ResourceType, Unknown]]
    point: Optional[float]


class GachaExtra(ModelWithExtra):
    id: Optional[int]
    resourceBoxId: Optional[int]


class GiftGachaExchange(ModelWithExtra):
    id: Optional[int]
    gachaId: Optional[int]
    seq: Optional[int]
    resourceBoxId: Optional[int]


class PracticeTicket(ModelWithExtra):
    id: Optional[int]
    name: Optional[str]
    exp: Optional[int]
    flavorText: Optional[str]


class SkillPracticeTicket(PracticeTicket):
    pass


class Level(ModelWithExtra):
    id: Optional[int]
    levelType: Optional[Union[LevelType, Unknown]]
    level: Optional[int]
    totalExp: Optional[int]


class Episode(ModelWithExtra):
    id: Optional[int]
    episodeNo: Optional[int]
    title: Optional[str]
    assetbundleName: Optional[str]
    scenarioId: Optional[str]
    releaseConditionId: Optional[int]
    rewardResourceBoxIds: Optional[List[int]]


class UnitStoryEpisode(Episode):
    unit: Optional[Union[Unit, Unknown]]
    chapterNo: Optional[int]
    unitEpisodeCategory: Optional[Union[Unit, Unknown]]
    episodeNoLabel: Optional[str]
    limitedReleaseStartAt: Optional[datetime]
    limitedReleaseEndAt: Optional[datetime]
    andReleaseConditionId: Optional[int]


class Chapter(ModelWithExtra):
    id: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    chapterNo: Optional[int]
    title: Optional[str]
    assetbundleName: Optional[str]
    episodes: Optional[List[UnitStoryEpisode]]


class UnitStory(ModelWithExtra):
    unit: Optional[Union[Unit, Unknown]]
    seq: Optional[int]
    assetbundleName: Optional[str]
    chapters: Optional[List[Chapter]]


class SpecialStoryEpisode(Episode):
    specialStoryId: Optional[int]
    specialStoryEpisodeType: Optional[str]
    isAbleSkip: Optional[bool]
    isActionSetRefresh: Optional[bool]
    specialStoryEpisodeTypeId: Optional[int]


class SpecialStory(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    title: Optional[str]
    assetbundleName: Optional[str]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    episodes: Optional[List[SpecialStoryEpisode]]


class Config(ModelWithExtra):
    configKey: Optional[str]
    value: Optional[str]


class ClientConfig(ModelWithExtra):
    id: Optional[int]
    value: Optional[str]
    type: Optional[Union[ClientConfigType, Unknown]]


class Wording(ModelWithExtra):
    wordingKey: Optional[str]
    value: Optional[str]


class Costume3d(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    costume3dGroupId: Optional[int]
    costume3dType: Optional[Union[Costume3dType, Unknown]]
    name: Optional[str]
    partType: Optional[Union[PartType, Unknown]]
    colorId: Optional[int]
    colorName: Optional[str]
    characterId: Optional[int]
    costume3dRarity: Optional[Union[Costume3dRarity, Unknown]]
    howToObtain: Optional[str]
    assetbundleName: Optional[str]
    designer: Optional[str]
    archiveDisplayType: Optional[Union[ArchiveDisplayType, Unknown]]
    archivePublishedAt: Optional[datetime]
    publishedAt: Optional[datetime]


class Costume3dModel(ModelWithExtra):
    id: Optional[int]
    costume3dId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    headCostume3dAssetbundleType: Optional[Union[HeadCostume3dAssetbundleType, Unknown]]
    thumbnailAssetbundleName: Optional[str]
    assetbundleName: Optional[str]
    colorAssetbundleName: Optional[str]
    part: Optional[str]


class Costume3dModelAvailablePattern(ModelWithExtra):
    id: Optional[int]
    headCostume3dId: Optional[int]
    hairCostume3dId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    isDefault: Optional[bool]


class GameCharacterUnit3dMotion(ModelWithExtra):
    id: Optional[int]
    gameCharacterUnitId: Optional[int]
    motionType: Optional[Union[MotionType, Unknown]]
    assetbundleName: Optional[str]


class Costume2d(ModelWithExtra):
    id: Optional[int]
    costume2dGroupId: Optional[int]
    character2dId: Optional[int]
    fromMmddhh: Optional[str]
    toMmddhh: Optional[str]
    live2dAssetbundleName: Optional[str]
    spineAssetbundleName: Optional[str]


class Costume2dGroup(ModelWithExtra):
    id: Optional[int]
    name: Optional[str]


class Topic(ModelWithExtra):
    id: Optional[int]
    topicType: Optional[Union[TopicType, Unknown]]
    topicTypeId: Optional[int]
    releaseConditionId: Optional[int]


class LiveStage(ModelWithExtra):
    id: Optional[int]
    name: Optional[str]
    assetbundleName: Optional[str]


class Stamp(ModelWithExtra):
    id: Optional[int]
    stampType: Optional[Union[StampType, Unknown]]
    seq: Optional[int]
    name: Optional[str]
    assetbundleName: Optional[str]
    balloonAssetbundleName: Optional[str]
    characterId1: Optional[int]
    gameCharacterUnitId: Optional[int]
    archivePublishedAt: Optional[datetime]
    description: Optional[str]
    archiveDisplayType: Optional[Union[ArchiveDisplayType, Unknown]]


class MultiLiveLobby(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    photonLobbyName: Optional[str]
    matchingLogic: Optional[Union[MatchingLogic, Unknown]]
    totalPower: Optional[int]
    assetbundleName: Optional[str]
    multiLiveLobbyType: Optional[Union[MultiLiveLobbyType, Unknown]]


class MasterLessonCost(Cost):
    id: Optional[int]
    cardRarityType: Optional[Union[CardRarityType, Unknown]]
    masterRank: Optional[int]
    seq: Optional[int]


class MasterLesson(ModelWithExtra):
    cardRarityType: Optional[Union[CardRarityType, Unknown]]
    masterRank: Optional[int]
    power1BonusFixed: Optional[int]
    power2BonusFixed: Optional[int]
    power3BonusFixed: Optional[int]
    characterRankExp: Optional[int]
    costs: Optional[List[MasterLessonCost]]
    rewards: Optional[List[Union[dict, str, int]]]


class MasterLessonReward(ModelWithExtra):
    id: Optional[int]
    cardId: Optional[int]
    masterRank: Optional[int]
    seq: Optional[int]
    resourceBoxId: Optional[int]
    cardRarity: Optional[int]


class CardExchangeResource(ModelWithExtra):
    cardRarityType: Optional[Union[CardRarityType, Unknown]]
    seq: Optional[int]
    resourceBoxId: Optional[int]


class MaterialExchangeCost(Cost):
    materialExchangeId: Optional[int]
    costGroupId: Optional[int]
    seq: Optional[int]


class MaterialExchange(ModelWithExtra):
    id: Optional[int]
    materialExchangeSummaryId: Optional[int]
    seq: Optional[int]
    resourceBoxId: Optional[int]
    refreshCycle: Optional[Union[RefreshCycle, Unknown]]
    costs: Optional[List[MaterialExchangeCost]]
    exchangeLimit: Optional[int]
    startAt: Optional[datetime]


class MaterialExchangeSummary(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    exchangeCategory: Optional[Union[ExchangeCategory, Unknown]]
    materialExchangeType: Optional[Union[MaterialExchangeType, Unknown]]
    name: Optional[str]
    assetbundleName: Optional[str]
    startAt: Optional[datetime]
    materialExchanges: Optional[List[MaterialExchange]]
    endAt: Optional[datetime]
    notificationRemainHour: Optional[int]


class BoostItem(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    recoveryValue: Optional[int]
    assetBundleName: Optional[str]
    flavorText: Optional[str]


class BillingProduct(ModelWithExtra):
    id: Optional[int]
    groupId: Optional[int]
    platform: Optional[Platform]
    productId: Optional[str]
    price: Optional[int]
    unitPrice: Optional[float]


class BillingShopItem(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    billingShopItemType: Optional[Union[BillingShopItemType, Unknown]]
    billingProductGroupId: Optional[int]
    name: Optional[str]
    description: Optional[str]
    billableLimitType: Optional[Union[BillableLimitType, Unknown]]
    billableLimitResetIntervalType: Optional[Union[BillableLimitResetIntervalType, Unknown]]
    assetbundleName: Optional[str]
    resourceBoxId: Optional[int]
    billableLimitValue: Optional[int]
    bonusResourceBoxId: Optional[int]
    label: Optional[str]
    endAt: Optional[datetime]
    startAt: Optional[datetime]
    billableLimitResetIntervalValue: Optional[int]
    purchaseOption: Optional[Union[PurchaseOption, Unknown]]


class BillingShopItemExchangeCost(Cost):
    id: Optional[int]
    billingShopItemId: Optional[int]


class ColorfulPass(ModelWithExtra):
    id: Optional[int]
    resourceBoxId: Optional[int]
    receivableDays: Optional[int]
    presentSentence: Optional[str]
    expireDays: Optional[int]
    dailyPaidGachaSpinLimit: Optional[int]
    challengeLivePointMultiple: Optional[int]
    livePointMultiple: Optional[int]


class JewelBehavior(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    jewelBehaviorType: Optional[Union[JewelBehaviorType, Unknown]]
    jewelBehaviorTypeQuantity: Optional[int]
    amount: Optional[int]


class CharacterRankAchieveResource(ModelWithExtra):
    releaseConditionId: Optional[int]
    characterId: Optional[int]
    characterRank: Optional[int]
    resources: Optional[List[Union[dict, str, int]]]


class CharacterRank(ModelWithExtra):
    id: Optional[int]
    characterId: Optional[int]
    characterRank: Optional[int]
    power1BonusRate: Optional[float]
    power2BonusRate: Optional[float]
    power3BonusRate: Optional[float]
    rewardResourceBoxIds: Optional[List[int]]
    characterRankAchieveResources: Optional[List[CharacterRankAchieveResource]]


class CharacterMissionV2(ModelWithExtra):
    id: Optional[int]
    characterMissionType: Optional[Union[CharacterMissionType, Unknown]]
    characterId: Optional[int]
    parameterGroupId: Optional[int]
    sentence: Optional[str]
    progressSentence: Optional[str]
    isAchievementMission: Optional[bool]


class CharacterMissionV2ParameterGroup(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    requirement: Optional[int]
    exp: Optional[int]


class CharacterMissionV2AreaItem(ModelWithExtra):
    id: Optional[int]
    characterMissionType: Optional[Union[CharacterMissionType, Unknown]]
    areaItemId: Optional[int]
    characterId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]


class SystemLive2d(Live2d):
    serif: Optional[str]
    voice: Optional[str]
    publishedAt: Optional[datetime]
    closedAt: Optional[datetime]
    specialSeasonId: Optional[int]


class Reward(ModelWithExtra):
    id: Optional[int]
    missionType: Optional[Union[MissionType, Unknown]]
    missionId: Optional[int]
    seq: Optional[int]
    resourceBoxId: Optional[int]


class NormalMission(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    normalMissionType: Optional[Union[NormalMissionType, Unknown]]
    requirement: Optional[int]
    sentence: Optional[str]
    rewards: Optional[List[Reward]]


class BeginnerMission(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    beginnerMissionType: Optional[Union[BeginnerMissionType, Unknown]]
    beginnerMissionCategory: Optional[Union[BeginnerMissionCategory, Unknown]]
    requirement: Optional[int]
    sentence: Optional[str]
    rewards: Optional[List[Reward]]


class Detail(ModelWithExtra):
    resourceBoxPurpose: Optional[Union[ResourceBoxPurpose, Unknown]]
    resourceBoxId: Optional[int]
    seq: Optional[int]
    resourceType: Optional[Union[ResourceType, Unknown]]
    resourceQuantity: Optional[int]
    resourceId: Optional[int]
    resourceLevel: Optional[int]


class ResourceBox(ModelWithExtra):
    resourceBoxPurpose: Optional[Union[ResourceBoxPurpose, Unknown]]
    id: Optional[int]
    resourceBoxType: Optional[Union[ResourceBoxType, Unknown]]
    details: Optional[List[Detail]]
    description: Optional[str]
    name: Optional[str]
    assetbundleName: Optional[str]


class LiveMissionPeriod(ModelWithExtra):
    id: Optional[int]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    sentence: Optional[str]


class LiveMission(ModelWithExtra):
    id: Optional[int]
    liveMissionPeriodId: Optional[int]
    liveMissionType: Optional[Union[LiveMissionType, Unknown]]
    requirement: Optional[int]
    rewards: Optional[List[Reward]]


class LiveMissionPass(ModelWithExtra):
    id: Optional[int]
    liveMissionPeriodId: Optional[int]
    costumeName: Optional[str]
    character3dId1: Optional[int]
    character3dId2: Optional[int]
    maleAssetbundleName: Optional[str]
    femaleAssetbundleName: Optional[str]


class PenlightColor(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    description: Optional[str]
    colorCode: Optional[str]
    characterId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]


class Penlight(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    defaultPenlightColorId: Optional[int]
    assetbundleName: Optional[str]


class LiveTalk(ModelWithExtra):
    id: Optional[int]
    liveTalkType: Optional[Union[LiveTalkType, Unknown]]
    scenarioId: Optional[str]
    characterId1: Optional[int]
    characterId2: Optional[int]


class Tip(ModelWithExtra):
    id: Optional[int]
    title: Optional[str]
    description: Optional[str]
    fromUserRank: Optional[int]
    toUserRank: Optional[int]
    assetbundleName: Optional[str]


class GachaCeilItem(ModelWithExtra):
    id: Optional[int]
    gachaId: Optional[int]
    name: Optional[str]
    assetbundleName: Optional[str]
    convertStartAt: Optional[datetime]
    convertResourceBoxId: Optional[int]


class GachaCeilExchangeCost(Cost):
    gachaCeilExchangeId: Optional[int]
    gachaCeilItemId: Optional[int]


class GachaCeilExchangeSubstituteCost(Cost):
    id: Optional[int]
    substituteQuantity: Optional[int]


class GachaCeilExchange(ModelWithExtra):
    id: Optional[int]
    gachaCeilExchangeSummaryId: Optional[int]
    seq: Optional[int]
    resourceBoxId: Optional[int]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    gachaCeilExchangeCost: Optional[GachaCeilExchangeCost]
    gachaCeilExchangeSubstituteCosts: Optional[List[GachaCeilExchangeSubstituteCost]]
    exchangeLimit: Optional[int]
    gachaCeilExchangeLabelType: Optional[Union[GachaCeilExchangeLabelType, Unknown]]
    substituteLimit: Optional[int]


class GachaCeilExchangeSummary(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    gachaId: Optional[int]
    assetbundleName: Optional[str]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    gachaCeilExchanges: Optional[List[GachaCeilExchange]]
    gachaCeilItemId: Optional[int]


class PlayerRankReward(ModelWithExtra):
    id: Optional[int]
    playerRank: Optional[int]
    seq: Optional[int]
    resourceBoxId: Optional[int]


class GachaTicket(ModelWithExtra):
    id: Optional[int]
    name: Optional[str]
    assetbundleName: Optional[str]


class HonorGroup(ModelWithExtra):
    id: Optional[int]
    name: Optional[str]
    honorType: Optional[Union[HonorType, Unknown]]
    archivePublishedAt: Optional[datetime]
    backgroundAssetbundleName: Optional[str]


class HonorLevel(ModelWithExtra):
    honorId: Optional[int]
    level: Optional[int]
    bonus: Optional[int]
    description: Optional[str]


class Honor(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    groupId: Optional[int]
    honorRarity: Optional[Union[HonorRarity, Unknown]]
    name: Optional[str]
    assetbundleName: Optional[str]
    levels: Optional[List[HonorLevel]]
    honorTypeId: Optional[int]


class HonorMission(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    honorMissionType: Optional[Union[HonorMissionType, Unknown]]
    requirement: Optional[int]
    sentence: Optional[str]
    rewards: Optional[List[Reward]]


class BondsHonorLevel(ModelWithExtra):
    id: Optional[int]
    bondsHonorId: Optional[int]
    level: Optional[int]
    description: Optional[str]


class BondsHonor(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    bondsGroupId: Optional[int]
    gameCharacterUnitId1: Optional[int]
    gameCharacterUnitId2: Optional[int]
    honorRarity: Optional[Union[HonorRarity, Unknown]]
    name: Optional[str]
    description: Optional[str]
    levels: Optional[List[BondsHonorLevel]]
    configurableUnitVirtualSinger: Optional[bool]


class BondsHonorWord(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    bondsGroupId: Optional[int]
    assetbundleName: Optional[str]
    name: Optional[str]
    description: Optional[str]


class BondsReward(ModelWithExtra):
    id: Optional[int]
    bondsGroupId: Optional[int]
    rank: Optional[int]
    seq: Optional[int]
    bondsRewardType: Optional[Union[BondsRewardType, Unknown]]
    resourceBoxId: Optional[int]
    description: Optional[str]


class ChallengeLiveDetail(ModelWithExtra):
    id: Optional[int]
    challengeLiveId: Optional[int]
    challengeLiveType: Optional[Union[LiveType, Unknown]]


class ChallengeLive(ModelWithExtra):
    id: Optional[int]
    playableCount: Optional[int]
    challengeLiveDetails: Optional[List[ChallengeLiveDetail]]


class ChallengeLiveDeck(ModelWithExtra):
    id: Optional[int]
    characterId: Optional[int]
    releaseConditionId: Optional[int]
    cardLimit: Optional[int]


class ChallengeLiveStage(ModelWithExtra):
    id: Optional[int]
    characterId: Optional[int]
    rank: Optional[int]
    name: Optional[str]
    nextStageChallengePoint: Optional[int]
    completeStageResourceBoxId: Optional[int]
    completeStageCharacterExp: Optional[int]


class ChallengeLiveStageEx(ModelWithExtra):
    id: Optional[int]
    characterId: Optional[int]
    fromRank: Optional[int]
    toRank: Optional[int]
    name: Optional[str]
    nextStageChallengePoint: Optional[int]
    completeStageResourceBoxId: Optional[int]
    completeStageCharacterExp: Optional[int]


class ChallengeLiveHighScoreReward(ModelWithExtra):
    id: Optional[int]
    characterId: Optional[int]
    highScore: Optional[int]
    resourceBoxId: Optional[int]


class ChallengeLiveCharacter(ModelWithExtra):
    id: Optional[int]
    characterId: Optional[int]
    releaseConditionId: Optional[int]
    orReleaseConditionId: Optional[int]


class ChallengeLivePlayDayReward(ModelWithExtra):
    id: Optional[int]
    challengeLivePlayDayRewardPeriodId: Optional[int]
    playDays: Optional[int]
    seq: Optional[int]
    resourceBoxId: Optional[int]


class ChallengeLivePlayDayRewardPeriod(ModelWithExtra):
    id: Optional[int]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    priority: Optional[int]
    challengeLivePlayDayRewards: Optional[List[ChallengeLivePlayDayReward]]


class VirtualLiveSetlist(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    seq: Optional[int]
    virtualLiveSetlistType: Optional[Union[VirtualLiveSetlistType, Unknown]]
    assetbundleName: Optional[str]
    virtualLiveStageId: Optional[int]
    musicId: Optional[int]
    musicVocalId: Optional[int]
    character3dId1: Optional[int]
    character3dId2: Optional[int]
    character3dId3: Optional[int]
    character3dId4: Optional[int]
    character3dId5: Optional[int]
    character3dId6: Optional[int]


class VirtualLiveBeginnerSchedule(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    dayOfWeek: Optional[Union[DayOfWeek, Unknown]]
    startTime: Optional[str]
    endTime: Optional[str]


class VirtualLiveSchedule(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    seq: Optional[int]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    noticeGroupId: Optional[str]


class VirtualLiveCharacter(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    gameCharacterUnitId: Optional[int]
    seq: Optional[int]


class VirtualLiveReward(ModelWithExtra):
    id: Optional[int]
    virtualLiveType: Optional[Union[VirtualLiveType, Unknown]]
    virtualLiveId: Optional[int]
    resourceBoxId: Optional[int]


class VirtualLiveWaitingRoom(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    assetbundleName: Optional[str]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    lobbyAssetbundleName: Optional[str]


class VirtualItem(ModelWithExtra):
    id: Optional[int]
    virtualItemCategory: Optional[Union[VirtualItemCategory, Unknown]]
    seq: Optional[int]
    priority: Optional[int]
    name: Optional[str]
    assetbundleName: Optional[str]
    costVirtualCoin: Optional[int]
    costJewel: Optional[int]
    cheerPoint: Optional[int]
    effectAssetbundleName: Optional[str]
    effectExpressionType: Optional[Union[EffectExpressionType, Unknown]]
    unit: Optional[Union[Unit, Unknown]]
    gameCharacterUnitId: Optional[int]
    virtualItemLabelType: Optional[Union[VirtualItemLabelType, Unknown]]


class VirtualLiveAppeal(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    virtualLiveStageStatus: Optional[Union[VirtualLiveStageStatus, Unknown]]
    appealText: Optional[str]


class VirtualLiveInformation(ModelWithExtra):
    virtualLiveId: Optional[int]
    summary: Optional[str]
    description: Optional[str]


class VirtualLive(ModelWithExtra):
    id: Optional[int]
    virtualLiveType: Optional[Union[VirtualLiveType, Unknown]]
    virtualLivePlatform: Optional[str]
    seq: Optional[int]
    name: Optional[str]
    assetbundleName: Optional[str]
    screenMvMusicVocalId: Optional[int]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    rankingAnnounceAt: Optional[datetime]
    virtualLiveSetlists: Optional[List[VirtualLiveSetlist]]
    virtualLiveBeginnerSchedules: Optional[List[VirtualLiveBeginnerSchedule]]
    virtualLiveSchedules: Optional[List[VirtualLiveSchedule]]
    virtualLiveCharacters: Optional[List[VirtualLiveCharacter]]
    virtualLiveReward: Optional[VirtualLiveReward]
    virtualLiveRewards: Optional[List[VirtualLiveReward]]
    virtualLiveCheerPointRewards: Optional[List[Union[dict, str, int]]]
    virtualLiveWaitingRoom: Optional[VirtualLiveWaitingRoom]
    virtualItems: Optional[List[VirtualItem]]
    virtualLiveAppeals: Optional[List[VirtualLiveAppeal]]
    virtualLiveInformation: Optional[VirtualLiveInformation]
    archiveReleaseConditionId: Optional[int]
    virtualLiveTicketId: Optional[int]


class VirtualShopItem(ModelWithExtra):
    id: Optional[int]
    virtualShopId: Optional[int]
    virtualShopItemType: Optional[Union[VirtualShopItemType, Unknown]]
    seq: Optional[int]
    resourceBoxId: Optional[int]
    costVirtualCoin: Optional[int]
    costJewel: Optional[int]
    startAt: Optional[datetime]
    costPaidJewel: Optional[int]
    endAt: Optional[datetime]
    limitValue: Optional[int]


class VirtualShop(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    virtualShopItems: Optional[List[VirtualShopItem]]
    virtualShopType: Optional[Union[VirtualShopType, Unknown]]
    virtualLiveId: Optional[int]


class VirtualLiveCheerMessage(ModelWithExtra):
    id: Optional[int]
    virtualLiveType: Optional[Union[VirtualLiveType, Unknown]]
    resourceType: Optional[Union[ResourceType, Unknown]]
    fromCostVirtualCoin: Optional[int]
    toCostVirtualCoin: Optional[int]
    fromCost: Optional[int]
    toCost: Optional[int]
    assetbundleName: Optional[str]
    messageLengthLimit: Optional[int]
    displaySec: Optional[float]
    messageSize: Optional[str]
    colorCode: Optional[str]
    virtualLiveCheerMessageDisplayLimitId: Optional[int]


class VirtualLiveCheerMessageDisplayLimit(ModelWithExtra):
    id: Optional[int]
    displayLimit: Optional[int]


class VirtualLiveTicket(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    virtualLiveTicketType: Optional[Union[VirtualLiveTicketType, Unknown]]
    name: Optional[str]
    flavorText: Optional[str]
    assetbundleName: Optional[str]


class VirtualLivePamphlet(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    name: Optional[str]
    flavorText: Optional[str]
    assetbundleName: Optional[str]


class AvatarAccessory(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    part: Optional[Union[AccessoryPart, Unknown]]
    assetbundleName: Optional[str]


class AvatarCostume(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    assetbundleName: Optional[str]


class AvatarMotion(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    syncMusicFlg: Optional[bool]
    repeatCount: Optional[int]
    assetbundleName: Optional[str]


class AvatarSkinColor(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    colorCode: Optional[str]


class AvatarCoordinate(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    assetbundleName: Optional[str]
    skinColorCode: Optional[str]
    costumeAssetbundleName: Optional[str]
    accessoryPart: Optional[Union[AccessoryPart, Unknown]]
    accessoryAssetbundleName: Optional[str]


class NgWord(ModelWithExtra):
    id: Optional[int]
    word: Optional[str]


class RuleSlide(ModelWithExtra):
    id: Optional[int]
    ruleSlideType: Optional[Union[RuleSlideType, Unknown]]
    assetbundleName: Optional[str]


class Facility(ModelWithExtra):
    id: Optional[int]
    facilityType: Optional[Union[FacilityType, Unknown]]
    releaseConditionId: Optional[int]
    andReleaseConditionId: Optional[int]


class OneTimeBehavior(ModelWithExtra):
    id: Optional[int]
    oneTimeBehaviorType: Optional[Union[OneTimeBehaviorType, Unknown]]
    releaseConditionId: Optional[int]


class LoginBonus(ModelWithExtra):
    id: Optional[int]
    day: Optional[int]
    resourceBoxId: Optional[int]


class BeginnerLoginBonus(ModelWithExtra):
    id: Optional[int]
    day: Optional[int]
    resourceBoxId: Optional[int]
    loginBonusId: Optional[int]


class BeginnerLoginBonusSummary(ModelWithExtra):
    id: Optional[int]
    loginBonusId: Optional[int]
    startAt: Optional[datetime]
    endAt: Optional[datetime]


class LimitedLoginBonusDetail(ModelWithExtra):
    id: Optional[int]
    limitedLoginBonusId: Optional[int]
    day: Optional[int]
    resourceBoxId: Optional[int]


class LimitedLoginBonus(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    assetbundleName: Optional[str]
    closeAt: Optional[datetime]
    limitedLoginBonusDetails: Optional[List[LimitedLoginBonusDetail]]


class LoginBonusLive2d(Live2d):
    serif: Optional[str]
    voice: Optional[str]
    publishedAt: Optional[datetime]
    closedAt: Optional[datetime]


class EventRankingReward(ModelWithExtra):
    id: Optional[int]
    eventRankingRewardRangeId: Optional[int]
    resourceBoxId: Optional[int]


class EventRankingRewardRange(ModelWithExtra):
    id: Optional[int]
    eventId: Optional[int]
    fromRank: Optional[int]
    toRank: Optional[int]
    eventRankingRewards: Optional[List[EventRankingReward]]


class Event(ModelWithExtra):
    id: Optional[int]
    eventType: Optional[Union[EventType, Unknown]]
    name: Optional[str]
    assetbundleName: Optional[str]
    bgmAssetbundleName: Optional[str]
    startAt: Optional[datetime]
    aggregateAt: Optional[datetime]
    rankingAnnounceAt: Optional[datetime]
    distributionStartAt: Optional[datetime]
    closedAt: Optional[datetime]
    distributionEndAt: Optional[datetime]
    virtualLiveId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    eventRankingRewardRanges: Optional[List[EventRankingRewardRange]]
    eventPointAssetbundleName: Optional[str]


class EventMusic(ModelWithExtra):
    eventId: Optional[int]
    musicId: Optional[int]
    seq: Optional[int]
    releaseConditionId: Optional[int]


class EventDeckBonus(ModelWithExtra):
    id: Optional[int]
    eventId: Optional[int]
    gameCharacterUnitId: Optional[int]
    cardAttr: Optional[Union[CardAttr, Unknown]]
    bonusRate: Optional[float]


class EventRarityBonusRate(ModelWithExtra):
    id: Optional[int]
    cardRarityType: Optional[Union[CardRarityType, Unknown]]
    masterRank: Optional[int]
    bonusRate: Optional[float]


class EventItem(ModelWithExtra):
    id: Optional[int]
    eventId: Optional[int]
    name: Optional[str]
    flavorText: Optional[str]
    assetbundleName: Optional[str]


class EpisodeReward(ModelWithExtra):
    storyType: Optional[Union[StoryType, Unknown]]
    resourceBoxId: Optional[int]


class EventStoryEpisode(ModelWithExtra):
    id: Optional[int]
    eventStoryId: Optional[int]
    episodeNo: Optional[int]
    title: Optional[str]
    assetbundleName: Optional[str]
    scenarioId: Optional[str]
    releaseConditionId: Optional[int]
    episodeRewards: Optional[List[EpisodeReward]]


class EventStory(ModelWithExtra):
    id: Optional[int]
    eventId: Optional[int]
    assetbundleName: Optional[str]
    eventStoryEpisodes: Optional[List[EventStoryEpisode]]


class EventExchangeCost(Cost):
    id: Optional[int]
    eventExchangeId: Optional[int]


class EventExchange(ModelWithExtra):
    id: Optional[int]
    eventExchangeSummaryId: Optional[int]
    seq: Optional[int]
    resourceBoxId: Optional[int]
    exchangeLimit: Optional[int]
    eventExchangeCost: Optional[EventExchangeCost]


class EventExchangeSummary(ModelWithExtra):
    id: Optional[int]
    eventId: Optional[int]
    assetbundleName: Optional[str]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    eventExchanges: Optional[List[EventExchange]]


class EventStoryUnit(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    eventStoryId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    eventStoryUnitRelation: Optional[Union[EventStoryUnitRelation, Unknown]]


class EventCard(ModelWithExtra):
    id: Optional[int]
    cardId: Optional[int]
    eventId: Optional[int]
    bonusRate: Optional[float]


class PreliminaryTournamentCard(ModelWithExtra):
    id: Optional[int]
    preliminaryTournamentId: Optional[int]
    cardId: Optional[int]


class PreliminaryTournamentMusic(ModelWithExtra):
    id: Optional[int]
    preliminaryTournamentId: Optional[int]
    musicDifficultyId: Optional[int]
    musicId: Optional[int]


class PreliminaryTournament(ModelWithExtra):
    id: Optional[int]
    preliminaryTournamentType: Optional[Union[PreliminaryTournamentType, Unknown]]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    releaseConditionId: Optional[int]
    preliminaryTournamentCards: Optional[List[PreliminaryTournamentCard]]
    preliminaryTournamentMusics: Optional[List[PreliminaryTournamentMusic]]


class CheerfulCarnivalSummary(ModelWithExtra):
    id: Optional[int]
    eventId: Optional[int]
    theme: Optional[str]
    midtermAnnounce1At: Optional[datetime]
    midtermAnnounce2At: Optional[datetime]
    assetbundleName: Optional[str]


class CheerfulCarnivalTeam(ModelWithExtra):
    id: Optional[int]
    eventId: Optional[int]
    seq: Optional[int]
    teamName: Optional[str]
    assetbundleName: Optional[str]


class CheerfulCarnivalPartyName(ModelWithExtra):
    id: Optional[int]
    partyName: Optional[str]
    gameCharacterUnitId1: Optional[int]
    gameCharacterUnitId2: Optional[int]
    gameCharacterUnitId3: Optional[int]
    gameCharacterUnitId4: Optional[int]
    gameCharacterUnitId5: Optional[int]


class CheerfulCarnivalCharacterPartyName(ModelWithExtra):
    id: Optional[int]
    characterPartyName: Optional[str]
    gameCharacterUnitId: Optional[int]


class CheerfulCarnivalLiveTeamPointBonus(ModelWithExtra):
    id: Optional[int]
    teamPointBonusRate: Optional[int]


class CheerfulCarnivalReward(ModelWithExtra):
    id: Optional[int]
    eventId: Optional[int]
    cheerfulCarnivalTeamId: Optional[int]
    resourceBoxId: Optional[int]


class CheerfulCarnivalResultReward(ModelWithExtra):
    id: Optional[int]
    eventId: Optional[int]
    cheerfulCarnivalTeamPointTermType: Optional[Union[CheerfulCarnivalTeamPointTermType, Unknown]]
    cheerfulCarnivalResultType: Optional[Union[CheerfulCarnivalResultType, Unknown]]
    resourceBoxId: Optional[int]


class Appeal(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    appealTargetType: Optional[Union[AppealTargetType, Unknown]]
    appealType: Optional[Union[AppealType, Unknown]]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    appealReadConditionType: Optional[Union[AppealReadConditionType, Unknown]]
    text: Optional[str]


class Boost(ModelWithExtra):
    id: Optional[int]
    costBoost: Optional[int]
    isEventOnly: Optional[bool]
    expRate: Optional[int]
    rewardRate: Optional[int]
    livePointRate: Optional[int]
    eventPointRate: Optional[int]
    bondsExpRate: Optional[int]


class BoostPresent(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    recoveryValue: Optional[int]
    presentLimit: Optional[int]
    assetbundleName: Optional[str]
    isUnlimitedReceive: Optional[bool]
    boostPresentCostId: Optional[int]


class BoostPresentCost(Cost):
    id: Optional[int]


class EpisodeCharacter(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    character2dId: Optional[int]
    storyType: Optional[Union[StoryType, Unknown]]
    episodeId: Optional[int]


class CustomProfileTextColor(ModelWithExtra):
    id: Optional[int]
    seq: Optional[int]
    colorCode: Optional[str]


class CustomProfileTextFont(ModelWithExtra):
    id: Optional[int]
    name: Optional[str]
    fontName: Optional[str]
    assetbundleName: Optional[str]


class CustomProfileResource(ModelWithExtra):
    customProfileResourceType: Optional[Union[CustomProfileResourceType, Unknown]]
    id: Optional[int]
    seq: Optional[int]
    name: Optional[str]
    resourceLoadType: Optional[Union[ResourceLoadType, Unknown]]
    resourceLoadVal: Optional[str]
    fileName: Optional[str]


class CustomProfilePlayerInfoResource(CustomProfileResource):
    pass


class CustomProfileGeneralBackgroundResource(CustomProfileResource):
    pass


class CustomProfileStoryBackgroundResource(CustomProfileResource):
    pass


class CustomProfileCollectionResource(CustomProfileResource):
    customProfileResourceCollectionType: Optional[Union[CustomProfileResourceCollectionType, Unknown]]
    groupId: Optional[int]


class CustomProfileMemberStandingPictureResource(CustomProfileResource):
    characterId: Optional[int]


class CustomProfileShapeResource(CustomProfileResource):
    pass


class CustomProfileEtcResource(CustomProfileResource):
    pass


class CustomProfileGachaBehavior(ModelWithExtra):
    id: Optional[int]
    customProfileGachaId: Optional[int]
    seq: Optional[int]
    costResourceType: Optional[Union[ResourceType, Unknown]]
    costResourceQuantity: Optional[int]
    spinCount: Optional[int]


class CustomProfileGachaDetail(ModelWithExtra):
    id: Optional[int]
    customProfileGachaId: Optional[int]
    customProfileResourceType: Optional[Union[CustomProfileResourceType, Unknown]]
    customProfileResourceId: Optional[int]
    customProfileResourceQuantity: Optional[int]
    weight: Optional[int]


class CustomProfileGacha(ModelWithExtra):
    id: Optional[int]
    name: Optional[str]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    assetbundleName: Optional[str]
    description: Optional[str]
    notice: Optional[str]
    customProfileGachaBehaviors: Optional[List[CustomProfileGachaBehavior]]
    customProfileGachaDetails: Optional[List[CustomProfileGachaDetail]]


class StreamingLiveBgm(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    seq: Optional[int]
    musicVocalId: Optional[int]


class Omikuji(ModelWithExtra):
    id: Optional[int]
    omikujiGroupId: Optional[int]
    unit: Optional[Union[Unit, Unknown]]
    fortuneType: Optional[Union[FortuneType, Unknown]]
    summary: Optional[str]
    title1: Optional[str]
    description1: Optional[str]
    title2: Optional[str]
    description2: Optional[str]
    title3: Optional[str]
    description3: Optional[str]
    unitAssetbundleName: Optional[str]
    fortuneAssetbundleName: Optional[str]
    omikujiCoverAssetbundleName: Optional[str]
    unitFilePath: Optional[str]
    fortuneFilePath: Optional[str]
    omikujiCoverFilePath: Optional[str]


class OmikujiGroup(ModelWithExtra):
    id: Optional[int]
    name: Optional[str]
    summary: Optional[str]
    description: Optional[str]
    assetbundleName: Optional[str]
    appealAssetbundleName: Optional[str]
    soundAssetbundleName: Optional[str]


class OmikujiRate(ModelWithExtra):
    id: Optional[int]
    omikujiGroupId: Optional[int]
    fortuneType: Optional[Union[FortuneType, Unknown]]
    rate: Optional[float]


class OmikujiCost(Cost):
    id: Optional[int]
    omikujiGroupId: Optional[int]
    seq: Optional[int]


class OmikujiReward(ModelWithExtra):
    id: Optional[int]
    omikujiGroupId: Optional[int]
    seq: Optional[int]
    resourceType: Optional[Union[ResourceType, Unknown]]
    resourceId: Optional[int]
    resourceQuantity: Optional[int]


class VirtualBoothShop(ModelWithExtra):
    id: Optional[int]
    virtualLiveId: Optional[int]
    virtualBoothShopType: Optional[Union[VirtualBoothShopType, Unknown]]
    targetId: Optional[int]


class SpecialSeason(ModelWithExtra):
    id: Optional[int]
    specialSeasonType: Optional[Union[SpecialSeasonType, Unknown]]
    startAt: Optional[datetime]
    endAt: Optional[datetime]
    priority: Optional[int]


class SpecialSeasonArea(ModelWithExtra):
    id: Optional[int]
    specialSeasonId: Optional[int]
    areaId: Optional[int]
    assetbundleName: Optional[str]
    fileName: Optional[str]
    specialSeasonAreaUseType: Optional[Union[SpecialSeasonAreaUseType, Unknown]]


class RankMatchPenalty(ModelWithExtra):
    id: Optional[int]
    count: Optional[int]
    rankMatchPenaltyType: Optional[Union[RankMatchPenaltyType, Unknown]]
    rankMatchPenaltyTypeValue: Optional[int]


class RankMatchPlacement(ModelWithExtra):
    id: Optional[int]
    rankMatchPlacementConditionType: Optional[str]
    tierBehaviorType: Optional[Union[TierBehaviorType, Unknown]]
    tierBehaviorTypeValue: Optional[int]
    rankMatchPlacementConditionTypeValue: Optional[int]


class RankMatchBonusPointCondition(ModelWithExtra):
    id: Optional[int]
    rankMatchBonusPointConditionType: Optional[Union[RankMatchBonusPointConditionType, Unknown]]
    groupId: Optional[int]
    priority: Optional[int]
    calcType: Optional[Union[CalcType, Unknown]]
    bonusPoint: Optional[int]


class RankMatchSeasonPlayableTime(ModelWithExtra):
    id: Optional[int]
    rankMatchSeasonId: Optional[int]
    startTime: Optional[str]
    endTime: Optional[str]


class RankMatchSeasonTierMusicPlayLevel(ModelWithExtra):
    id: Optional[int]
    rankMatchSeasonId: Optional[int]
    rankMatchTierId: Optional[int]
    fromPlayLevel: Optional[int]
    toPlayLevel: Optional[int]


class RankMatchSeasonTierReward(ModelWithExtra):
    id: Optional[int]
    rankMatchSeasonId: Optional[int]
    rankMatchTierId: Optional[int]
    resourceBoxId: Optional[int]


class RankMatchSeason(ModelWithExtra):
    id: Optional[int]
    name: Optional[str]
    startAt: Optional[datetime]
    aggregatedAt: Optional[datetime]
    rankingPublishedAt: Optional[datetime]
    batchExecutionAt: Optional[datetime]
    distributionStartAt: Optional[datetime]
    distributionEndAt: Optional[datetime]
    closedAt: Optional[datetime]
    assetbundleName: Optional[str]
    isDisplayResult: Optional[bool]
    rankMatchSeasonPlayableTimes: Optional[List[RankMatchSeasonPlayableTime]]
    rankMatchSeasonTierMusicPlayLevels: Optional[
        List[RankMatchSeasonTierMusicPlayLevel]
    ]
    rankMatchSeasonTierRewards: Optional[List[RankMatchSeasonTierReward]]


class RankMatchTier(ModelWithExtra):
    id: Optional[int]
    rankMatchGradeId: Optional[int]
    rankMatchClassId: Optional[int]
    tier: Optional[int]
    totalMusicDifficulty: Optional[int]
    point: Optional[int]
    gradeAssetbundleName: Optional[str]
    tierAssetbundleName: Optional[str]


class RankMatchTierBonusPoint(ModelWithExtra):
    id: Optional[int]
    rankMatchTierId: Optional[int]
    maxBonusPoint: Optional[int]
    rewardPoint: Optional[int]


class RankMatchGrade(Costume2dGroup):
    id: Optional[int]
    name: Optional[str]


class RankMatchClass(Costume2dGroup):
    id: Optional[int]
    name: Optional[str]


class LimitedTitleScreen(ModelWithExtra):
    id: Optional[int]
    priority: Optional[int]
    downloadStartAt: Optional[datetime]
    downloadEndAt: Optional[datetime]
    displayStartAt: Optional[datetime]
    displayEndAt: Optional[datetime]
    bgAssetbundleName: Optional[str]
    logoAssetbundleName: Optional[str]
    bgmAssetbundleName: Optional[str]
    startEffectAssetbundleName: Optional[str]


class MasterData(ModelWithExtra):
    gameCharacters: Optional[List[GameCharacter]]
    gameCharacterUnits: Optional[List[GameCharacterUnit]]
    outsideCharacters: Optional[List[OutsideCharacter]]
    character3ds: Optional[List[Character3d]]
    character2ds: Optional[List[Character2d]]
    characterProfiles: Optional[List[CharacterProfile]]
    bonds: Optional[List[Bond]]
    bondsLive2ds: Optional[List[BondsLive2d]]
    bondsRankUpLive2ds: Optional[List[BondsRankUpLive2d]]
    unitProfiles: Optional[List[UnitProfile]]
    actionSets: Optional[List[ActionSet]]
    areas: Optional[List[Area]]
    areaPlaylists: Optional[List[AreaPlaylist]]
    mobCharacters: Optional[List[MobCharacter]]
    characterCostumes: Optional[List[CharacterCostume]]
    cardCostume3ds: Optional[List[CardCostume3d]]
    cards: Optional[List[Card]]
    skills: Optional[List[Skill]]
    cardEpisodes: Optional[List[CardEpisode]]
    cardRarities: Optional[List[CardRarity]]
    cardSkillCosts: Optional[List[CardSkillCost]]
    musics: Optional[List[Music]]
    musicTags: Optional[List[MusicTag]]
    musicDifficulties: Optional[List[MusicDifficulty]]
    musicVocals: Optional[List[MusicVocal]]
    musicDanceMembers: Optional[List[MusicDanceMember]]
    musicAchievements: Optional[List[MusicAchievement]]
    musicVideoCharacters: Optional[List[MusicVideoCharacter]]
    musicAssetVariants: Optional[List[MusicAssetVariant]]
    musicCollaborations: Optional[List[MusicCollaboration]]
    episodeMusicVideoCostumes: Optional[List[EpisodeMusicVideoCostume]]
    releaseConditions: Optional[List[ReleaseCondition]]
    playLevelScores: Optional[List[PlayLevelScore]]
    ingameCombos: Optional[List[IngameCombo]]
    ingameNotes: Optional[List[IngameNote]]
    ingameNoteJudges: Optional[List[IngameNoteJudge]]
    ingamePlayLevels: Optional[List[IngamePlayLevel]]
    ingameCutins: Optional[List[IngameCutin]]
    ingameCutinCharacters: Optional[List[IngameCutinCharacter]]
    ingameJudgeFrames: Optional[List[IngameJudgeFrame]]
    ingameNoteJudgeTechnicalScores: Optional[List[IngameNoteJudgeTechnicalScore]]
    shops: Optional[List[Shop]]
    shopItems: Optional[List[ShopItem]]
    costume3dShopItems: Optional[List[Costume3dShopItem]]
    areaItems: Optional[List[AreaItem]]
    areaItemLevels: Optional[List[AreaItemLevel]]
    materials: Optional[List[Material]]
    gachas: Optional[List[Gacha]]
    gachaBonuses: Optional[List[GachaBonus]]
    gachaBonusPoints: Optional[List[GachaBonusPoint]]
    gachaExtras: Optional[List[GachaExtra]]
    giftGachaExchanges: Optional[List[GiftGachaExchange]]
    gachaTabs: Optional[List[Union[dict, str, int]]]
    practiceTickets: Optional[List[PracticeTicket]]
    skillPracticeTickets: Optional[List[SkillPracticeTicket]]
    levels: Optional[List[Level]]
    unitStories: Optional[List[UnitStory]]
    specialStories: Optional[List[SpecialStory]]
    configs: Optional[List[Config]]
    clientConfigs: Optional[List[ClientConfig]]
    wordings: Optional[List[Wording]]
    costume3ds: Optional[List[Costume3d]]
    costume3dModels: Optional[List[Costume3dModel]]
    costume3dModelAvailablePatterns: Optional[List[Costume3dModelAvailablePattern]]
    gameCharacterUnit3dMotions: Optional[List[GameCharacterUnit3dMotion]]
    costume2ds: Optional[List[Costume2d]]
    costume2dGroups: Optional[List[Costume2dGroup]]
    topics: Optional[List[Topic]]
    liveStages: Optional[List[LiveStage]]
    stamps: Optional[List[Stamp]]
    multiLiveLobbies: Optional[List[MultiLiveLobby]]
    masterLessons: Optional[List[MasterLesson]]
    masterLessonRewards: Optional[List[MasterLessonReward]]
    cardExchangeResources: Optional[List[CardExchangeResource]]
    materialExchanges: Optional[List[MaterialExchange]]
    materialExchangeSummaries: Optional[List[MaterialExchangeSummary]]
    boostItems: Optional[List[BoostItem]]
    billingProducts: Optional[List[BillingProduct]]
    billingShopItems: Optional[List[BillingShopItem]]
    billingShopItemExchangeCosts: Optional[List[BillingShopItemExchangeCost]]
    colorfulPasses: Optional[List[ColorfulPass]]
    jewelBehaviors: Optional[List[JewelBehavior]]
    characterRanks: Optional[List[CharacterRank]]
    characterMissionV2s: Optional[List[CharacterMissionV2]]
    characterMissionV2ParameterGroups: Optional[List[CharacterMissionV2ParameterGroup]]
    characterMissionV2AreaItems: Optional[List[CharacterMissionV2AreaItem]]
    systemLive2ds: Optional[List[SystemLive2d]]
    normalMissions: Optional[List[NormalMission]]
    beginnerMissions: Optional[List[BeginnerMission]]
    resourceBoxes: Optional[List[ResourceBox]]
    liveMissionPeriods: Optional[List[LiveMissionPeriod]]
    liveMissions: Optional[List[LiveMission]]
    liveMissionPasses: Optional[List[LiveMissionPass]]
    penlightColors: Optional[List[PenlightColor]]
    penlights: Optional[List[Penlight]]
    liveTalks: Optional[List[LiveTalk]]
    tips: Optional[List[Tip]]
    gachaCeilItems: Optional[List[GachaCeilItem]]
    gachaCeilExchangeSummaries: Optional[List[GachaCeilExchangeSummary]]
    playerRankRewards: Optional[List[PlayerRankReward]]
    gachaTickets: Optional[List[GachaTicket]]
    honorGroups: Optional[List[HonorGroup]]
    honors: Optional[List[Honor]]
    honorMissions: Optional[List[HonorMission]]
    bondsHonors: Optional[List[BondsHonor]]
    bondsHonorWords: Optional[List[BondsHonorWord]]
    bondsRewards: Optional[List[BondsReward]]
    challengeLives: Optional[List[ChallengeLive]]
    challengeLiveDecks: Optional[List[ChallengeLiveDeck]]
    challengeLiveStages: Optional[List[ChallengeLiveStage]]
    challengeLiveStageExs: Optional[List[ChallengeLiveStageEx]]
    challengeLiveHighScoreRewards: Optional[List[ChallengeLiveHighScoreReward]]
    challengeLiveCharacters: Optional[List[ChallengeLiveCharacter]]
    challengeLivePlayDayRewardPeriods: Optional[List[ChallengeLivePlayDayRewardPeriod]]
    virtualLives: Optional[List[VirtualLive]]
    virtualShops: Optional[List[VirtualShop]]
    virtualItems: Optional[List[VirtualItem]]
    virtualLiveCheerMessages: Optional[List[VirtualLiveCheerMessage]]
    virtualLiveCheerMessageDisplayLimits: Optional[
        List[VirtualLiveCheerMessageDisplayLimit]
    ]
    virtualLiveTickets: Optional[List[VirtualLiveTicket]]
    virtualLivePamphlets: Optional[List[VirtualLivePamphlet]]
    avatarAccessories: Optional[List[AvatarAccessory]]
    avatarCostumes: Optional[List[AvatarCostume]]
    avatarMotions: Optional[List[AvatarMotion]]
    avatarSkinColors: Optional[List[AvatarSkinColor]]
    avatarCoordinates: Optional[List[AvatarCoordinate]]
    ngWords: Optional[List[NgWord]]
    ruleSlides: Optional[List[RuleSlide]]
    facilities: Optional[List[Facility]]
    oneTimeBehaviors: Optional[List[OneTimeBehavior]]
    loginBonuses: Optional[List[LoginBonus]]
    beginnerLoginBonuses: Optional[List[BeginnerLoginBonus]]
    beginnerLoginBonusSummaries: Optional[List[BeginnerLoginBonusSummary]]
    limitedLoginBonuses: Optional[List[LimitedLoginBonus]]
    loginBonusLive2ds: Optional[List[LoginBonusLive2d]]
    events: Optional[List[Event]]
    eventMusics: Optional[List[EventMusic]]
    eventDeckBonuses: Optional[List[EventDeckBonus]]
    eventRarityBonusRates: Optional[List[EventRarityBonusRate]]
    eventItems: Optional[List[EventItem]]
    eventStories: Optional[List[EventStory]]
    eventExchangeSummaries: Optional[List[EventExchangeSummary]]
    eventStoryUnits: Optional[List[EventStoryUnit]]
    eventCards: Optional[List[EventCard]]
    preliminaryTournaments: Optional[List[PreliminaryTournament]]
    cheerfulCarnivalSummaries: Optional[List[CheerfulCarnivalSummary]]
    cheerfulCarnivalTeams: Optional[List[CheerfulCarnivalTeam]]
    cheerfulCarnivalPartyNames: Optional[List[CheerfulCarnivalPartyName]]
    cheerfulCarnivalCharacterPartyNames: Optional[
        List[CheerfulCarnivalCharacterPartyName]
    ]
    cheerfulCarnivalLiveTeamPointBonuses: Optional[
        List[CheerfulCarnivalLiveTeamPointBonus]
    ]
    cheerfulCarnivalRewards: Optional[List[CheerfulCarnivalReward]]
    cheerfulCarnivalResultRewards: Optional[List[CheerfulCarnivalResultReward]]
    appeals: Optional[List[Appeal]]
    boosts: Optional[List[Boost]]
    boostPresents: Optional[List[BoostPresent]]
    boostPresentCosts: Optional[List[BoostPresentCost]]
    episodeCharacters: Optional[List[EpisodeCharacter]]
    customProfileTextColors: Optional[List[CustomProfileTextColor]]
    customProfileTextFonts: Optional[List[CustomProfileTextFont]]
    customProfilePlayerInfoResources: Optional[List[CustomProfilePlayerInfoResource]]
    customProfileGeneralBackgroundResources: Optional[
        List[CustomProfileGeneralBackgroundResource]
    ]
    customProfileStoryBackgroundResources: Optional[
        List[CustomProfileStoryBackgroundResource]
    ]
    customProfileCollectionResources: Optional[List[CustomProfileCollectionResource]]
    customProfileMemberStandingPictureResources: Optional[
        List[CustomProfileMemberStandingPictureResource]
    ]
    customProfileShapeResources: Optional[List[CustomProfileShapeResource]]
    customProfileEtcResources: Optional[List[CustomProfileEtcResource]]
    customProfileMemberResourceExcludeCards: Optional[List[Union[dict, str, int]]]
    customProfileGachas: Optional[List[CustomProfileGacha]]
    customProfileGachaTabs: Optional[List[Union[dict, str, int]]]
    streamingLiveBgms: Optional[List[StreamingLiveBgm]]
    omikujis: Optional[List[Omikuji]]
    omikujiGroups: Optional[List[OmikujiGroup]]
    omikujiRates: Optional[List[OmikujiRate]]
    omikujiCosts: Optional[List[OmikujiCost]]
    omikujiRewards: Optional[List[OmikujiReward]]
    virtualBoothShops: Optional[List[VirtualBoothShop]]
    specialSeasons: Optional[List[SpecialSeason]]
    specialSeasonAreas: Optional[List[SpecialSeasonArea]]
    rankMatchPenalties: Optional[List[RankMatchPenalty]]
    rankMatchPlacements: Optional[List[RankMatchPlacement]]
    rankMatchBonusPointConditions: Optional[List[RankMatchBonusPointCondition]]
    rankMatchSeasons: Optional[List[RankMatchSeason]]
    rankMatchTiers: Optional[List[RankMatchTier]]
    rankMatchTierBonusPoints: Optional[List[RankMatchTierBonusPoint]]
    rankMatchGrades: Optional[List[RankMatchGrade]]
    rankMatchClasses: Optional[List[RankMatchClass]]
    limitedTitleScreens: Optional[List[LimitedTitleScreen]]
    panelMissionCampaigns: Optional[List[Union[dict, str, int]]]



# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

class ProjectSekaiException(Exception):
    pass

class NotAuthenticatedException(ProjectSekaiException):
    pass

class ServerInMaintenance(ProjectSekaiException):
    pass

class NoAvailableVersions(ProjectSekaiException):
    pass

class SessionExpired(ProjectSekaiException):
    pass

class UpdateRequired(ProjectSekaiException):
    pass

class AppUpdateRequired(UpdateRequired):
    def __init__(self, appVersion: str, appHash: str, multiPlayVersion: str):
        self.appVersion = appVersion
        self.appHash = appHash
        self.multiPlayVersion = multiPlayVersion

class DataUpdateRequired(UpdateRequired):
    def __init__(self, dataVersion: str, appVersionStatus: str):
        self.dataVersion = dataVersion
        self.appVersionStatus = appVersionStatus

class AssetUpdateRequired(UpdateRequired):
    def __init__(self, assetVersion: str, assetHash: str):
        self.assetVersion = assetVersion
        self.assetHash = assetHash

class MultipleUpdatesRequired(UpdateRequired):
    def __init__(self, dataVersion: str, assetVersion: str, assetHash: str, appVersionStatus: str):
        self.dataVersion = dataVersion
        self.assetVersion = assetVersion
        self.assetHash = assetHash
        self.appVersionStatus = appVersionStatus

class TutorialEnded(ProjectSekaiException):
    pass

class LiveActive(ProjectSekaiException):
    pass

class LiveNotActive(ProjectSekaiException):
    pass

class LiveDead(ProjectSekaiException):
    pass

class LiveNotDead(ProjectSekaiException):
    pass


# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from typing import Optional, Union

from .model import ModelWithExtra
from pjsekai.enums import *

class SystemInfo(ModelWithExtra):
    systemProfile: Optional[str]
    appVersion: Optional[str]
    multiPlayVersion: Optional[str]
    dataVersion: Optional[str]
    assetVersion: Optional[str]
    appHash: Optional[str]
    assetHash: Optional[str]
    appVersionStatus: Union[AppVersionStatus, Unknown, None]



# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from typing import List, Optional, Union, Dict

from .model import ModelWithExtra
from pjsekai.enums import *


class Bundle(ModelWithExtra):
    bundleName: Optional[str]
    cacheDirectoryName: Optional[str]
    hash: Optional[str]
    category: Union[BundleCategory, Unknown, None]
    crc: Optional[int]
    fileSize: Optional[int]
    dependencies: Optional[List[str]]
    paths: Optional[List[str]]
    isBuiltin: Optional[bool]


class AssetBundleInfo(ModelWithExtra):
    version: Optional[str]
    os: Union[AssetOS, Unknown, None]
    bundles: Optional[Dict[str, Bundle]]



# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from .enums import *
from .platform import *
from .tutorialStatus import *
from .unknown import *


# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from typing import Optional, List
from json import load, dump, JSONDecodeError
from pydantic.json import pydantic_encoder

from pjsekai.models import *
from pjsekai.enums import *
from pjsekai.api import API
from pjsekai.asset import Asset
from pjsekai.exceptions import *
from pjsekai.utilities import *
from pjsekai.live import *


class Client:

    _key: bytes
    _iv: bytes
    _hcaKey: bytes
    _jwtSecret: str
    _platform: Platform
    _systemInfoFilePath: Optional[str]
    _masterDataFilePath: Optional[str]
    _userDataFilePath: Optional[str]
    _assetsPath: Optional[str]
    _asset: Optional[Asset]
    _systemInfo: SystemInfo
    _userId: Optional[str]
    _credential: Optional[str]
    _masterData: MasterData
    _userData: dict

    @property
    def platform(self) -> Platform:
        return self._platform
    @property
    def domains(self) -> Dict[str,str]:
        return self.apiManager.domains
    @property 
    def DEFAULT_DOMAINS(self) -> Dict[str,str]:
        return self.apiManager.DEFAULT_DOMAINS
    @property
    def enableEncryption(self) -> Dict[str,bool]:
        return self.apiManager.enableEncryption
    @property 
    def DEFAULT_ENCRYPTION(self) -> Dict[str,bool]:
        return self.apiManager.DEFAULT_ENCRYPTION
    @property
    def systemInfo(self) -> SystemInfo:
        return self._systemInfo
    @systemInfo.setter
    def systemInfo(self, newValue: SystemInfo) -> None:
        self._systemInfo = newValue
        if self._systemInfoFilePath is not None:
            with open(self._systemInfoFilePath, "w") as f:
                dump(newValue,f,indent=2,ensure_ascii=False,default=pydantic_encoder)
    @property
    def userId(self) -> Optional[str]:
        return self._userId
    @property
    def credential(self) -> Optional[str]:
        return self._credential
    @property
    def isLoggedIn(self) -> bool:
        return self._credential is not None
    @property
    def masterData(self) -> MasterData:
        return self._masterData
    @masterData.setter
    def masterData(self, newValue: MasterData) -> None:
        self._masterData = newValue
        if self._masterDataFilePath is not None:
            with open(self._masterDataFilePath, "w") as f:
                dump(newValue,f,indent=2,ensure_ascii=False,default=pydantic_encoder)
    @property
    def userData(self) -> dict:
        return self._userData 
    def _updateUserResources(self, response) -> dict:
        self._userData = {
            **self._userData,
            **response["updatedResources"],
        }
        if self._userDataFilePath is not None:
            with open(self._userDataFilePath, "w") as f:
                dump(self._userData,f,indent=2,ensure_ascii=False)
        del response["updatedResources"]
        return response
    @property
    def asset(self) -> Optional[Asset]:
        return self._asset

    @property
    def now(self) -> Optional[int]:
        return self._userData["now"]
    @property
    def friends(self) -> Optional[List[dict]]:
        if not self.isLoggedIn:
            raise NotAuthenticatedException()
        if "userFriends" not in self._userData:
            return None
        return [friend for friend in self._userData["userFriends"] if friend["friendStatus"]=="friend"]
    @property
    def receivedFriendRequests(self) -> Optional[List[dict]]:
        if not self.isLoggedIn:
            raise NotAuthenticatedException()
        if "userFriends" not in self._userData:
            return None
        return [friend for friend in self._userData["userFriends"] if friend["friendStatus"]=="pending_request"]
    @property
    def sentFriendRequests(self) -> Optional[List[dict]]:
        if not self.isLoggedIn:
            raise NotAuthenticatedException()
        if "userFriends" not in self._userData:
            return None
        return [friend for friend in self._userData["userFriends"] if friend["friendStatus"]=="sent_request"]

    def __init__(
        self, 
        key: bytes, 
        iv: bytes, 
        hcaKey: bytes = b"", 
        jwtSecret: str = "", 
        platform: Platform = Platform.ANDROID, 
        domains: Optional[Dict[str,str]] = None, 
        systemInfoFilePath: Optional[str] = None, 
        masterDataFilePath: Optional[str] = None, 
        userDataFilePath: Optional[str] = None, 
        assetsPath: Optional[str] = None, 
        ensureLatest: bool = False,
        enableEncryption: Optional[Dict[str,bool]] = None,
    ) -> None:
        self._key = key
        self._iv = iv
        self._hcaKey = hcaKey
        self._jwtSecret = jwtSecret
        self._platform = platform
        
        self._systemInfoFilePath = systemInfoFilePath
        self._masterDataFilePath = masterDataFilePath
        self._userDataFilePath = userDataFilePath
        self._assetsPath = assetsPath
        self._asset = None

        if self._systemInfoFilePath is not None:
            try:
                with open(self._systemInfoFilePath, "r") as f:
                    self._systemInfo = SystemInfo(**load(f))
            except (FileNotFoundError, JSONDecodeError):
                self.systemInfo = SystemInfo()
        else:
            self.systemInfo = SystemInfo()

        if self._masterDataFilePath is not None:
            try:
                with open(self._masterDataFilePath, "r") as f:
                    self._masterData = MasterData(**load(f))
            except (FileNotFoundError, JSONDecodeError):
                self.masterData = MasterData()
        else:
            self.masterData = MasterData()

        self._userData = {}
        if self._userDataFilePath is not None:
            try:
                with open(self._userDataFilePath, "r") as f:
                    self._userData = load(f)
            except (FileNotFoundError, JSONDecodeError):
                with open(self._userDataFilePath, "w") as f:
                    dump(self._userData,f,indent=2,ensure_ascii=False)

        self._userId = None
        self._credential = None
        self.apiManager = API(
            platform = platform, 
            domains = domains, 
            key = key, 
            iv = iv, 
            jwtSecret = jwtSecret, 
            systemInfo = self.systemInfo,
            enableEncryption = enableEncryption,
        )
        self.apiManager.getSignedCookie()

        if self.systemInfo.assetVersion is not None and self.systemInfo.assetHash is not None and assetsPath is not None:
            self._asset = Asset(self.systemInfo.assetVersion,self.systemInfo.assetHash,assetsPath)

        if ensureLatest:
            self.ensureLatest()

    def register(self) -> dict:
        response: dict = self.apiManager.register()
        return self._updateUserResources(response)

    def login(self, userId: str, credential: str) -> dict:
        response: dict = self.apiManager.authenticate(userId,credential)
        self._userId = userId
        self._credential = credential
        self.apiManager.getLoginBonus()
        self._userData = self.apiManager.getUserData()
        return self._updateUserResources(response)

    def reloadUserData(self, name: Optional[str] = None) -> dict:
        self._userData = self.apiManager.getUserData(name)
        return self._userData

    def checkVersion(self, bypassAvailability: bool = False) -> SystemInfo:
        response: dict = self.apiManager.getSystemInfo()
        appVersions = [appVersionInfo for appVersionInfo in (SystemInfo(**appVersion) for appVersion in response["appVersions"]) if bypassAvailability or appVersionInfo.appVersionStatus is not AppVersionStatus.NOT_AVAILABLE]
        if len(appVersions) > 0:
            matchingAppVersionInfo = [appVersionInfo for appVersionInfo in appVersions if appVersionInfo.appVersion == self.systemInfo.appVersion]
            if len(matchingAppVersionInfo) > 0:
                info: SystemInfo = matchingAppVersionInfo[-1]
                status: str = "" if info.appVersionStatus is None else info.appVersionStatus.value
                if info.systemProfile != self.systemInfo.systemProfile:
                    self.systemInfo = SystemInfo(
                        systemProfile = info.systemProfile, 
                        appVersion = self.systemInfo.appVersion, 
                        appHash = self.systemInfo.appHash,
                        multiPlayVersion = self.systemInfo.multiPlayVersion,
                        appVersionStatus = status,
                    )
                assetUpdateRequired: bool = info.assetVersion != self.systemInfo.assetVersion or self.asset is None or info.assetVersion != self.asset.version
                if not bypassAvailability and info.appVersionStatus is AppVersionStatus.MAINTENANCE:
                    raise ServerInMaintenance()
                if info.dataVersion is not None and info.assetVersion is not None and info.assetHash is not None and info.dataVersion != self.systemInfo.dataVersion and assetUpdateRequired:
                    raise MultipleUpdatesRequired(info.dataVersion,info.assetVersion,info.assetHash,status)
                elif info.assetVersion is not None and info.assetHash is not None and assetUpdateRequired:
                    raise AssetUpdateRequired(info.assetVersion,info.assetHash)
                elif info.dataVersion is not None and info.dataVersion != self.systemInfo.dataVersion:
                    raise DataUpdateRequired(info.dataVersion,status)
                return info
            elif appVersions[-1].appVersion is not None and appVersions[-1].appHash is not None and appVersions[-1].multiPlayVersion is not None:
                raise AppUpdateRequired(appVersions[-1].appVersion,appVersions[-1].appHash,appVersions[-1].multiPlayVersion)
        raise NoAvailableVersions()
        
    def updateApp(self, appVersion: Optional[str] = None, appHash: Optional[str] = None, multiPlayVersion: Optional[str] = None) -> None:
        if appVersion is None or appHash is None or multiPlayVersion is None:
            try:
                self.checkVersion()
            except AppUpdateRequired as e:
                appVersion = e.appVersion
                appHash = e.appHash
                multiPlayVersion = e.multiPlayVersion
            except (ServerInMaintenance, MultipleUpdatesRequired, AssetUpdateRequired, DataUpdateRequired):
                return
        self.systemInfo = self.systemInfo.copy(update={
            "appVersion": appVersion,
            "appHash": appHash,
            "multiPlayVersion": multiPlayVersion,
        })

    def updateData(self, dataVersion: str, appVersionStatus: str) -> None:
        response: dict = self.apiManager.getMasterData(dataVersion)
        self.masterData = MasterData(**response)
        self.systemInfo = self.systemInfo.copy(update={
            "dataVersion": dataVersion,
            "appVersionStatus": appVersionStatus,
        })

    def updateAsset(self, assetVersion: str, assetHash: str) -> None:
        self._asset = Asset(assetVersion,assetHash,self._assetsPath)
        self._asset.getAssetBundleInfo(self.apiManager)
        self.systemInfo = self.systemInfo.copy(update={
            "assetVersion": assetVersion,
            "assetHash": assetHash,
        })

    def ensureLatest(self) -> bool:
        try:
            self.checkVersion()
        except AppUpdateRequired as e:
            self.updateApp(e.appVersion, e.appHash, e.multiPlayVersion)
            self.apiManager.systemInfo = self.systemInfo
            self.ensureLatest()
            return True
        except MultipleUpdatesRequired as e:
            self.updateAsset(e.assetVersion, e.assetHash)
            self.apiManager.systemInfo = self.systemInfo
            self.updateData(e.dataVersion, e.appVersionStatus)
            self.apiManager.systemInfo = self.systemInfo
            return True
        except AssetUpdateRequired as e:
            self.updateAsset(e.assetVersion, e.assetHash)
            self.apiManager.systemInfo = self.systemInfo
            return True
        except DataUpdateRequired as e:
            self.updateData(e.dataVersion, e.appVersionStatus)
            self.apiManager.systemInfo = self.systemInfo
            return True
        return False


    def refreshSignedCookie(self) -> None:
        self.apiManager.getSignedCookie()

    def ping(self) -> dict:
        return self.apiManager.ping()

    def transferOut(self, password: str) -> dict:
        response: dict = self.apiManager.generateTransferCode(password)
        return self._updateUserResources(response)

    def transferCheck(self, transferCode: str, password: str) -> dict:
        response: dict = self.apiManager.checkTransferCode(transferCode, password)
        return response

    def transferIn(self, transferCode: str, password: str) -> dict:
        response: dict = self.apiManager.generateCredential(transferCode, password)
        userId: str = response["afterUserGamedata"]["userId"]
        credential: str = response["credential"]
        return self.login(userId,credential)

    def advanceTutorial(self, unit: Unit = Unit.LN) -> dict:
        currentTutorialStatus: TutorialStatus = TutorialStatus(self.userData["userTutorial"]["tutorialStatus"])
        response: dict = self.apiManager.setTutorialStatus(currentTutorialStatus.next(unit))
        return self._updateUserResources(response)

    def receivePresent(self, presentId) -> dict:
        response: dict = self.apiManager.receivePresents([presentId])
        return self._updateUserResources(response)

    def receiveAllPresents(self) -> dict:
        response: dict = self.apiManager.receivePresents([present["presentId"] for present in self.userData["userPresents"]])
        return self._updateUserResources(response)

    def gacha(self, gachaId: int, gachaBehaviorId: int) -> dict:
        response: dict = self.apiManager.gacha(gachaId,gachaBehaviorId)
        return self._updateUserResources(response)

    def startSoloLive(self, live: SoloLive):
        response: dict = self.apiManager.startSoloLive(live.musicId, live.musicDifficultyId, live.musicVocalId, live.deckId, live.boostCount, live.isAuto)
        live.start(response["userLiveId"],response["skills"],response["comboCutins"])
        return self._updateUserResources(response)

    def endSoloLive(self, live: SoloLive) -> dict:
        if not live.isActive or live.liveId is None:
            raise LiveNotActive
        if live.life <= 0:
            raise LiveDead
        response: dict = self.apiManager.endSoloLive(live.liveId, live.score, live.perfectCount, live.greatCount, live.goodCount, live.badCount, live.missCount, live.maxCombo, live.life, live.tapCount, live.continueCount)
        live.end()
        return self._updateUserResources(response)

    def getEventRankings(self, eventId: int, targetUserId: Optional[str] = None, targetRank: Optional[int] = None, higherLimit: Optional[int] = None, lowerLimit: Optional[int] = None) -> dict:
        if targetUserId is None and targetRank is None:
            targetUserId = self.userId
        return self.apiManager.getEventRankings(eventId, targetUserId, targetRank, higherLimit, lowerLimit)

    def getEventTeamsPlayerCount(self, eventId: int) -> dict:
        return self.apiManager.getEventTeamsPlayerCount(eventId)

    def getEventTeamsPoint(self, eventId: int) -> dict:
        return self.apiManager.getEventTeamsPoint(eventId)

    def getRankMatchRankings(self, rankMatchSeasonId: int, targetUserId: Optional[str] = None, targetRank: Optional[int] = None, higherLimit: Optional[int] = None, lowerLimit: Optional[int] = None) -> dict:
        if targetUserId is None and targetRank is None:
            targetUserId = self.userId
        return self.apiManager.getRankMatchRankings(rankMatchSeasonId, targetUserId, targetRank, higherLimit, lowerLimit)

    def sendFriendRequest(self, userId: str, message: Optional[str] = None) -> None:
        response: dict = self.apiManager.sendFriendRequest(userId, message)
        self._updateUserResources(response)

    def rejectFriendRequest(self, requestUserId: str) -> None:
        response: dict = self.apiManager.rejectFriendRequest(requestUserId)
        self._updateUserResources(response)

    def acceptFriendRequest(self, requestUserId: str) -> dict:
        response: dict = self.apiManager.acceptFriendRequest(requestUserId)
        return self._updateUserResources(response)

    def removeFriend(self, friendUserId: str) -> dict:
        response: dict = self.apiManager.removeFriend(friendUserId)
        return self._updateUserResources(response)



# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from enum import Enum
from typing import Dict, List, Optional

from pjsekai.exceptions import LiveActive, LiveDead, LiveNotActive, LiveNotDead

class Judgement(Enum):
        PERFECT = "perfect"
        GREAT = "great"
        GOOD = "good"
        BAD = "bad"
        MISS = "miss"

        @property
        def willCombo(self) -> bool:
            return self is self.PERFECT or self is self.GREAT
        @property
        def competitiveScore(self) -> int:
            if self is self.PERFECT:
                return 3
            elif self is self.GREAT:
                return 2
            elif self is self.GOOD:
                return 1
            else:
                return 0
class SoloLive:

    _musicId: int
    _musicDifficultyId: int
    _musicVocalId: int
    _deckId: int
    _boostCount: int
    _isAuto: bool
    _isActive: bool
    _liveId: Optional[str]
    _skills: List[dict]
    _cutins: List[dict]
    _score: int
    _competitiveScore: int
    _judgementCounts: Dict[Judgement,int]
    _combo: int
    _maxCombo: int
    _life: int
    _tapCount: int
    _continueCount: int

    @property
    def musicId(self) -> int:
        return self._musicId
    @property
    def musicDifficultyId(self) -> int:
        return self._musicDifficultyId
    @property
    def musicVocalId(self) -> int:
        return self._musicVocalId
    @property
    def deckId(self) -> int:
        return self._deckId
    @property
    def boostCount(self) -> int:
        return self._boostCount
    @property
    def isAuto(self) -> bool:
        return self._isAuto

    @property
    def isActive(self) -> bool:
        return self._isActive
    @property
    def liveId(self) -> Optional[str]:
        return self._liveId
    @property
    def skills(self) -> List[dict]:
        return self._skills
    @property
    def cutins(self) -> List[dict]:
        return self._cutins
    @property
    def score(self) -> int:
        return self._score
    @property
    def competitiveScore(self) -> int:
        return self._competitiveScore
    @property
    def perfectCount(self) -> int:
        return self._judgementCounts[Judgement.PERFECT]
    @property
    def greatCount(self) -> int:
        return self._judgementCounts[Judgement.GREAT]
    @property
    def goodCount(self) -> int:
        return self._judgementCounts[Judgement.GOOD]
    @property
    def badCount(self) -> int:
        return self._judgementCounts[Judgement.BAD]
    @property
    def missCount(self) -> int:
        return self._judgementCounts[Judgement.MISS]
    @property
    def combo(self) -> int:
        return self._combo
    @property
    def maxCombo(self) -> int:
        return self._maxCombo
    @property
    def life(self) -> int:
        return self._life
    @property
    def tapCount(self) -> int:
        return self._tapCount
    @property
    def continueCount(self) -> int:
        return self._continueCount

    def __init__(self, musicId: int, musicDifficultyId: int, musicVocalId: int, deckId: int, boostCount: int = 0, isAuto: bool = False) -> None:
        self._musicId = musicId
        self._musicDifficultyId = musicDifficultyId
        self._musicVocalId = musicVocalId
        self._deckId = deckId
        self._boostCount = boostCount
        self._isAuto = isAuto

        self._liveId = None
        self._skills = []
        self._cutins = []
        self._score = 0
        self._competitiveScore = 0
        self._judgementCounts = {
            Judgement.PERFECT: 0,
            Judgement.GREAT: 0,
            Judgement.GOOD: 0,
            Judgement.BAD: 0,
            Judgement.MISS: 0,
        }
        self._combo = 0
        self._maxCombo = 0
        self._life = 1000
        self._tapCount = 0
        self._continueCount = 0

        self._isActive = False

    def start(self, userLiveId, skills, comboCutins) -> None:
        if self._isActive:
            raise LiveActive()
        self._liveId = userLiveId
        self._skills = skills
        self._cutins = comboCutins
        self._score = 0
        self._competitiveScore = 0
        self._judgementCounts = {
            Judgement.PERFECT: 0,
            Judgement.GREAT: 0,
            Judgement.GOOD: 0,
            Judgement.BAD: 0,
            Judgement.MISS: 0,
        }
        self._combo = 0
        self._maxCombo = 0
        self._life = 1000
        self._tapCount = 0
        self._continueCount = 0

        self._isActive = True
    
    def end(self) -> None:
        if not self._isActive:
            raise LiveNotActive()
        self._isActive = False

    def judge(self, judgement: Optional[Judgement] = None, isTap: bool = True, scoreChange: int = 0, lifeChange: int = 0) -> None:
        if not self._isActive:
            raise LiveNotActive()
        if self._life <= 0:
            raise LiveDead()
        if isTap:
            self._tapCount = self._tapCount + 1
        if judgement is not None:
            self._judgementCounts[judgement] = self._judgementCounts[judgement] + 1
            self._competitiveScore = self._competitiveScore + judgement.competitiveScore
            if judgement.willCombo:
                self._combo = self._combo + 1
                self._maxCombo = max(self._combo, self._maxCombo)
        self._score = self._score + scoreChange
        self._life = min(max(0,self._life + lifeChange),2000)

    def revive(self) -> None:
        if not self._isActive:
            raise LiveNotActive()
        if self._life > 0:
            raise LiveNotDead()
        self._continueCount = self._continueCount + 1
        self._life = 1000



# SPDX-FileCopyrightText: 2022-present Erik Chan <erikchan002@gmail.com>
#
# SPDX-License-Identifier: MIT

from typing import Optional
from json import JSONDecodeError, dump, load
from os import path, makedirs
from pydantic.json import pydantic_encoder

from pjsekai.api import API
from pjsekai.models import *

class Asset:

    _path: Optional[str]
    _version: str
    _hash: str
    _assetBundleInfo: Optional[AssetBundleInfo]

    @property
    def path(self) -> Optional[str]:
        return self._path
    @property
    def version(self) -> str:
        return self._version
    @property
    def hash(self) -> str:
        return self._hash
    @property
    def assetBundleInfo(self) -> Optional[AssetBundleInfo]:
        return self._assetBundleInfo
    @assetBundleInfo.setter
    def assetBundleInfo(self, newValue: Optional[AssetBundleInfo]) -> None:
        self._assetBundleInfo = newValue
        if self._path is not None and newValue is not None:
            makedirs(self._path,exist_ok=True)
            with open(path.join(self._path,"AssetBundleInfo.json"), "w") as f:
                dump(newValue,f,indent=2,ensure_ascii=False,default=pydantic_encoder)

    def __init__(self, version: str, hash: str, assetsPath: Optional[str] = None) -> None:
        self._path = assetsPath

        self._version = version
        self._hash = hash

        if self._path is not None:
            try:
                with open(path.join(self._path,"AssetBundleInfo.json"), "r") as f:
                    self._assetBundleInfo = AssetBundleInfo(**load(f))
            except (FileNotFoundError, JSONDecodeError):
                self.assetBundleInfo = None
        else:
            self.assetBundleInfo = None

    def getAssetBundleInfo(self, apiManager: API) -> AssetBundleInfo:
        self.assetBundleInfo = AssetBundleInfo(**apiManager.getAssetBundleInfo(self._version))
        return self.assetBundleInfo
